# Copyright (c) 2018,2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Test the `geometry` module."""


import logging

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_array_equal
from scipy.spatial import Delaunay

from metpy.interpolate.geometry import (area, circumcenter, circumcircle_radius, dist_2,
                                        distance, find_local_boundary, find_natural_neighbors,
                                        find_nn_triangles_point, get_point_count_within_r,
                                        get_points_within_r, order_edges, triangle_area)

logging.getLogger('metpy.interpolate.geometry').setLevel(logging.ERROR)


def test_get_points_within_r():
    r"""Test get points within a radius function."""
    x = list(range(10))
    y = list(range(10))

    center = [1, 5]

    radius = 5

    matches = get_points_within_r(center, list(zip(x, y)), radius).T

    truth = [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]

    assert_array_almost_equal(truth, matches)


def test_get_point_count_within_r():
    r"""Test get point count within a radius function."""
    x = list(range(10))
    y = list(range(10))

    center1 = [1, 5]
    center2 = [12, 10]

    radius = 5

    count = get_point_count_within_r([center1, center2], list(zip(x, y)), radius)

    truth = np.array([5, 2])

    assert_array_almost_equal(truth, count)


def test_triangle_area():
    r"""Test area of triangle function."""
    pt0 = [0, 0]
    pt1 = [10, 10]
    pt2 = [10, 0]

    truth = 50.0

    t_area = triangle_area(pt0, pt1, pt2)

    assert_almost_equal(truth, t_area)

    # what if two points are the same? Its a line!
    pt0 = [0, 0]
    pt1 = [0, 0]
    pt2 = [10, 0]

    truth = 0

    t_area = triangle_area(pt0, pt1, pt2)

    assert_almost_equal(truth, t_area)


def test_dist_2():
    r"""Test squared distance function."""
    x0 = 0
    y0 = 0

    x1 = 10
    y1 = 10

    truth = 200

    dist2 = dist_2(x0, y0, x1, y1)

    assert_almost_equal(truth, dist2)


def test_distance():
    r"""Test distance function."""
    pt0 = [0, 0]
    pt1 = [10, 10]

    truth = 14.14213562373095

    dist = distance(pt0, pt1)

    assert_almost_equal(truth, dist)


def test_circumcircle_radius():
    r"""Test circumcircle radius function."""
    pt0 = [0, 0]
    pt1 = [10, 10]
    pt2 = [10, 0]

    cc_r = circumcircle_radius(pt0, pt1, pt2)

    truth = 7.07

    assert_almost_equal(truth, cc_r, decimal=2)


def test_circumcircle_radius_degenerate():
    """Test that circumcircle_radius handles a degenerate triangle."""
    pt0 = [0, 0]
    pt1 = [10, 10]
    pt2 = [0, 0]

    assert np.isnan(circumcircle_radius(pt0, pt1, pt2))


def test_circumcenter():
    r"""Test circumcenter function."""
    pt0 = [0, 0]
    pt1 = [10, 10]
    pt2 = [10, 0]

    cc = circumcenter(pt0, pt1, pt2)

    truth = [5., 5.]

    assert_array_almost_equal(truth, cc)


def test_find_natural_neighbors():
    r"""Test find natural neighbors function."""
    x = list(range(0, 20, 4))
    y = list(range(0, 20, 4))
    gx, gy = np.meshgrid(x, y)
    pts = np.vstack([gx.ravel(), gy.ravel()]).T
    tri = Delaunay(pts)

    test_points = np.array([[2, 2], [5, 10], [12, 13.4], [12, 8], [20, 20]])

    neighbors, tri_info = find_natural_neighbors(tri, test_points)

    # Need to check point indices rather than simplex indices
    neighbors_truth = [[(1, 5, 0), (5, 1, 6)],
                       [(11, 17, 16), (17, 11, 12)],
                       [(23, 19, 24), (19, 23, 18), (23, 17, 18), (17, 23, 22)],
                       [(7, 13, 12), (13, 7, 8), (9, 13, 8), (13, 9, 14),
                        (13, 19, 18), (19, 13, 14), (17, 13, 18), (13, 17, 12)],
                       np.zeros((0, 3), dtype=np.int32)]

    centers_truth = [[(2.0, 2.0), (2.0, 2.0)],
                     [(6.0, 10.0), (6.0, 10.0)],
                     [(14.0, 14.0), (14.0, 14.0), (10.0, 14.0), (10.0, 14.0)],
                     [(10.0, 6.0), (10.0, 6.0), (14.0, 6.0), (14.0, 6.0),
                      (14.0, 10.0), (14.0, 10.0), (10.0, 10.0), (10.0, 10.0)],
                     np.zeros((0, 2), dtype=np.int32)]

    for i, true_neighbor in enumerate(neighbors_truth):
        assert set(true_neighbor) == {tuple(v) for v in tri.simplices[neighbors[i]]}
        assert set(centers_truth[i]) == {tuple(c) for c in tri_info[neighbors[i]]}


def test_find_nn_triangles_point():
    r"""Test find natural neighbors for a point function."""
    x = list(range(10))
    y = list(range(10))
    gx, gy = np.meshgrid(x, y)
    pts = np.vstack([gx.ravel(), gy.ravel()]).T
    tri = Delaunay(pts)

    tri_match = tri.find_simplex([4.5, 4.5])
    nn = find_nn_triangles_point(tri, tri_match, (4.5, 4.5))

    # Can't rely on simplex indices or the order of points that define them,
    # so need to be a bit more involved to check
    truth = [{45, 55, 44}, {55, 54, 44}]
    found = [set(verts) for verts in tri.simplices[nn]]
    for s in truth:
        assert s in found


def test_find_local_boundary():
    r"""Test find edges of natural neighbor triangle group function."""
    x = list(range(10))
    y = list(range(10))
    gx, gy = np.meshgrid(x, y)
    pts = np.vstack([gx.ravel(), gy.ravel()]).T
    tri = Delaunay(pts)

    tri_match = tri.find_simplex([4.5, 4.5])

    nn = find_nn_triangles_point(tri, tri_match, [4.5, 4.5])

    edges = find_local_boundary(tri, nn)

    truth = {(45, 55), (44, 45), (55, 54), (54, 44)}

    assert truth == set(edges)


def test_area():
    r"""Test get area of polygon function."""
    pt0 = [0, 0]
    pt1 = [5, 5]
    pt2 = [5, 0]

    truth = 12.5

    assert_almost_equal(area([pt0, pt1, pt2]), truth)


def test_order_edges():
    r"""Test order edges of polygon function."""
    edges = [[1, 2], [5, 6], [4, 5], [2, 3], [6, 1], [3, 4]]

    truth = [[1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 1]]

    assert_array_equal(truth, order_edges(edges))



# Copyright (c) 2017,2018 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tests for the `_util` module."""

from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr

from metpy.plots import add_metpy_logo, add_timestamp, add_unidata_logo, convert_gempak_color
from metpy.testing import autoclose_figure, get_test_data, version_check


@pytest.mark.mpl_image_compare(tolerance=2.638, remove_text=True)
def test_add_timestamp():
    """Test adding a timestamp to an axes object."""
    fig = plt.figure(figsize=(9, 9))
    ax = plt.subplot(1, 1, 1)
    add_timestamp(ax, time=datetime(2017, 1, 1))
    return fig


@pytest.mark.mpl_image_compare(tolerance=2.635, remove_text=True)
def test_add_timestamp_custom_format():
    """Test adding a timestamp to an axes object with custom time formatting."""
    fig = plt.figure(figsize=(9, 9))
    ax = plt.subplot(1, 1, 1)
    add_timestamp(ax, time=datetime(2017, 1, 1), time_format='%H:%M:%S %Y/%m/%d')
    return fig


@pytest.mark.mpl_image_compare(tolerance=5.389, remove_text=True)
def test_add_timestamp_pretext():
    """Test adding a timestamp to an axes object with custom pre-text."""
    fig = plt.figure(figsize=(9, 9))
    ax = plt.subplot(1, 1, 1)
    add_timestamp(ax, time=datetime(2017, 1, 1), pretext='Valid: ')
    return fig


@pytest.mark.mpl_image_compare(tolerance=0.844, remove_text=True)
def test_add_timestamp_high_contrast():
    """Test adding a timestamp to an axes object."""
    fig = plt.figure(figsize=(9, 9))
    ax = plt.subplot(1, 1, 1)
    add_timestamp(ax, time=datetime(2017, 1, 1), high_contrast=True)
    return fig


def test_add_timestamp_xarray():
    """Test that add_timestamp can work with xarray datetime accessor."""
    with autoclose_figure() as fig:
        ax = fig.add_subplot(1, 1, 1)
        ds = xr.open_dataset(get_test_data('AK-REGIONAL_8km_3.9_20160408_1445.gini'),
                             engine='gini')
        txt = add_timestamp(ax, ds.time.dt, pretext='')
        assert txt.get_text() == '2016-04-08T14:45:20Z'


@pytest.mark.mpl_image_compare(tolerance=0.004, remove_text=True)
def test_add_metpy_logo_small():
    """Test adding a MetPy logo to a figure."""
    fig = plt.figure(figsize=(9, 9))
    add_metpy_logo(fig)
    return fig


@pytest.mark.mpl_image_compare(tolerance=0.004, remove_text=True)
def test_add_metpy_logo_large():
    """Test adding a large MetPy logo to a figure."""
    fig = plt.figure(figsize=(9, 9))
    add_metpy_logo(fig, size='large')
    return fig


@pytest.mark.mpl_image_compare(tolerance=0.004, remove_text=True)
def test_add_unidata_logo():
    """Test adding a Unidata logo to a figure."""
    fig = plt.figure(figsize=(9, 9))
    add_unidata_logo(fig)
    return fig


def test_add_logo_invalid_size():
    """Test adding a logo to a figure with an invalid size specification."""
    with pytest.raises(ValueError), autoclose_figure(figsize=(9, 9)) as fig:
        add_metpy_logo(fig, size='jumbo')


@pytest.mark.mpl_image_compare(
    tolerance=1.072 if version_check('matplotlib<3.5') else 0,
    remove_text=True)
def test_gempak_color_image_compare():
    """Test creating a plot with all the GEMPAK colors."""
    c = range(32)
    mplc = convert_gempak_color(c)

    delta = 0.025
    x = y = np.arange(-3.0, 3.01, delta)
    xx, yy = np.meshgrid(x, y)
    z1 = np.exp(-xx**2 - yy**2)
    z2 = np.exp(-(xx - 1)**2 - (yy - 1)**2)
    z = (z1 - z2) * 2

    fig = plt.figure(figsize=(9, 9))
    cs = plt.contourf(xx, yy, z, levels=np.linspace(-1.8, 1.8, 33), colors=mplc)
    plt.colorbar(cs)
    return fig


@pytest.mark.mpl_image_compare(
    tolerance=1.215 if version_check('matplotlib<3.5') else 0,
    remove_text=True)
def test_gempak_color_xw_image_compare():
    """Test creating a plot with all the GEMPAK colors using xw style."""
    c = range(32)
    mplc = convert_gempak_color(c, style='xw')

    delta = 0.025
    x = y = np.arange(-3.0, 3.01, delta)
    xx, yy = np.meshgrid(x, y)
    z1 = np.exp(-xx**2 - yy**2)
    z2 = np.exp(-(xx - 1)**2 - (yy - 1)**2)
    z = (z1 - z2) * 2

    fig = plt.figure(figsize=(9, 9))
    cs = plt.contourf(xx, yy, z, levels=np.linspace(-1.8, 1.8, 33), colors=mplc)
    plt.colorbar(cs)
    return fig


def test_gempak_color_invalid_style():
    """Test converting a GEMPAK color with an invalid style parameter."""
    c = range(32)
    with pytest.raises(ValueError):
        convert_gempak_color(c, style='plt')


def test_gempak_color_quirks():
    """Test converting some unusual GEMPAK colors."""
    c = [-5, 95, 101]
    mplc = convert_gempak_color(c)
    truth = ['white', 'bisque', 'white']
    assert mplc == truth


def test_gempak_color_scalar():
    """Test converting a single GEMPAK color."""
    mplc = convert_gempak_color(6)
    truth = 'cyan'
    assert mplc == truth



# Copyright (c) 2021 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""
Custom flake8 plugin to catch MetPy-specific bad style/practice.

Currently this only looks for multiplication or division by units, since that can break
masked arrays and is slower than calling ``Quantity()``.
"""

import ast
from collections import namedtuple

Error = namedtuple('Error', 'lineno col code')


class MetPyVisitor(ast.NodeVisitor):
    """Visit nodes of the AST looking for violations."""

    def __init__(self):
        """Initialize the visitor."""
        self.errors = []

    @staticmethod
    def _is_unit(node):
        """Check whether a node should be considered to represent "units"."""
        # Looking for a .units attribute, a units.myunit, or a call to units()
        is_units_attr = isinstance(node, ast.Attribute) and node.attr == 'units'
        is_reg_attr = (isinstance(node, ast.Attribute)
                       and isinstance(node.value, ast.Name) and node.value.id == 'units')
        is_reg_call = (isinstance(node, ast.Call)
                       and isinstance(node.func, ast.Name) and node.func.id == 'units')
        is_unit_alias = isinstance(node, ast.Name) and 'unit' in node.id

        return is_units_attr or is_reg_attr or is_reg_call or is_unit_alias

    def visit_BinOp(self, node):  # noqa: N802
        """Visit binary operations."""
        # Check whether this is multiplying or dividing by units
        if (isinstance(node.op, (ast.Mult, ast.Div))
                and (self._is_unit(node.right) or self._is_unit(node.left))):
            self.error(node.lineno, node.col_offset, 1)

        super().generic_visit(node)

    def error(self, lineno, col, code):
        """Add an error to our output list."""
        self.errors.append(Error(lineno, col, code))


class MetPyChecker:
    """Flake8 plugin class to check MetPy style/best practice."""

    name = __name__
    version = '1.0'

    def __init__(self, tree):
        """Initialize the plugin."""
        self.tree = tree

    def run(self):
        """Run the plugin and yield errors."""
        visitor = MetPyVisitor()
        visitor.visit(self.tree)
        for err in visitor.errors:
            yield self.error(err)

    def error(self, err):
        """Format errors into Flake8's required format."""
        return (err.lineno, err.col,
                f'MPY{err.code:03d}: Multiplying/dividing by units--use units.Quantity()',
                type(self))



# Copyright (c) 2015,2016,2017,2018 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
r"""Provide tools for unit-aware, meteorological calculations."""

from .basic import *  # noqa: F403
from .cross_sections import *  # noqa: F403
from .exceptions import *  # noqa: F403
from .indices import *  # noqa: F403
from .kinematics import *  # noqa: F403
from .thermo import *  # noqa: F403
from .tools import *  # noqa: F403
from .turbulence import *  # noqa: F403
from ..package_tools import set_module

__all__ = basic.__all__[:]  # pylint: disable=undefined-variable
__all__.extend(cross_sections.__all__)  # pylint: disable=undefined-variable
__all__.extend(indices.__all__)  # pylint: disable=undefined-variable
__all__.extend(kinematics.__all__)  # pylint: disable=undefined-variable
__all__.extend(thermo.__all__)  # pylint: disable=undefined-variable
__all__.extend(tools.__all__)  # pylint: disable=undefined-variable
__all__.extend(turbulence.__all__)  # pylint: disable=undefined-variable

set_module(globals())



# Copyright (c) 2018,2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Test the `points` module."""


import logging

import numpy as np
import pytest
from scipy.spatial import cKDTree, Delaunay

from metpy.cbook import get_test_data
from metpy.interpolate import (interpolate_to_points, inverse_distance_to_points,
                               natural_neighbor_to_points)
from metpy.interpolate.geometry import dist_2, find_natural_neighbors
from metpy.interpolate.points import barnes_point, cressman_point, natural_neighbor_point
from metpy.testing import assert_almost_equal, assert_array_almost_equal
from metpy.units import units

logging.getLogger('metpy.interpolate.points').setLevel(logging.ERROR)


@pytest.fixture()
def test_data():
    r"""Return data used for tests in this file."""
    x = np.array([8, 67, 79, 10, 52, 53, 98, 34, 15, 58], dtype=float)
    y = np.array([24, 87, 48, 94, 98, 66, 14, 24, 60, 16], dtype=float)
    z = np.array([0.064, 4.489, 6.241, 0.1, 2.704, 2.809, 9.604, 1.156,
                  0.225, 3.364], dtype=float)

    return x, y, z


@pytest.fixture()
def test_points():
    r"""Return point locations used for tests in this file."""
    with get_test_data('interpolation_test_grid.npz') as fobj:
        data = np.load(fobj)
        return np.stack([data['xg'].reshape(-1), data['yg'].reshape(-1)], axis=1)


def test_nn_point(test_data):
    r"""Test find natural neighbors for a point interpolation function."""
    xp, yp, z = test_data

    tri = Delaunay(list(zip(xp, yp)))

    sim_gridx = [30]
    sim_gridy = [30]

    members, tri_info = find_natural_neighbors(tri,
                                               list(zip(sim_gridx, sim_gridy)))

    val = natural_neighbor_point(xp, yp, z, (sim_gridx[0], sim_gridy[0]),
                                 tri, members[0], tri_info)

    truth = 1.009

    assert_almost_equal(truth, val, 3)


def test_cressman_point(test_data):
    r"""Test Cressman interpolation for a point function."""
    xp, yp, z = test_data

    r = 40

    obs_tree = cKDTree(list(zip(xp, yp)))

    indices = obs_tree.query_ball_point([30, 30], r=r)

    dists = dist_2(30, 30, xp[indices], yp[indices])
    values = z[indices]

    truth = 1.05499444404

    value = cressman_point(dists, values, r)

    assert_almost_equal(truth, value)


def test_barnes_point(test_data):
    r"""Test Barnes interpolation for a point function."""
    xp, yp, z = test_data

    r = 40

    obs_tree = cKDTree(list(zip(xp, yp)))

    indices = obs_tree.query_ball_point([60, 60], r=r)

    dists = dist_2(60, 60, xp[indices], yp[indices])
    values = z[indices]

    assert_almost_equal(barnes_point(dists, values, 5762.7), 4.0871824)


def test_natural_neighbor_to_points(test_data, test_points):
    r"""Test natural neighbor interpolation to grid function."""
    xp, yp, z = test_data
    obs_points = np.vstack([xp, yp]).transpose()

    img = natural_neighbor_to_points(obs_points, z, test_points)

    with get_test_data('nn_bbox0to100.npz') as fobj:
        truth = np.load(fobj)['img'].reshape(-1)

    assert_array_almost_equal(truth, img)


def test_inverse_distance_to_points_invalid(test_data, test_points):
    """Test that inverse_distance_to_points raises when given an invalid method."""
    xp, yp, z = test_data
    obs_points = np.vstack([xp, yp]).transpose()
    with pytest.raises(ValueError):
        inverse_distance_to_points(obs_points, z, test_points, kind='shouldraise', r=40)


@pytest.mark.parametrize('assume_units', [None, 'mbar'])
@pytest.mark.parametrize('method', ['cressman', 'barnes'])
def test_inverse_distance_to_points(method, assume_units, test_data, test_points):
    r"""Test inverse distance interpolation to points function."""
    xp, yp, z = test_data
    obs_points = np.vstack([xp, yp]).transpose()

    extra_kw, test_file = {'cressman': ({'r': 20, 'min_neighbors': 1}, 'cressman_r20_mn1.npz'),
                           'barnes': ({'r': 40, 'kappa': 100}, 'barnes_r40_k100.npz')}[method]

    with get_test_data(test_file) as fobj:
        truth = np.load(fobj)['img'].reshape(-1)

    if assume_units:
        z = units.Quantity(z, assume_units)
        truth = units.Quantity(truth, assume_units)

    img = inverse_distance_to_points(obs_points, z, test_points, kind=method, **extra_kw)
    assert_array_almost_equal(truth, img)


def test_interpolate_to_points_invalid(test_data):
    """Test that interpolate_to_points raises when given an invalid method."""
    xp, yp, z = test_data
    obs_points = np.vstack([xp, yp]).transpose() * 10

    with get_test_data('interpolation_test_points.npz') as fobj:
        test_points = np.load(fobj)['points']

    with pytest.raises(ValueError):
        interpolate_to_points(obs_points, z, test_points, interp_type='shouldraise')


@pytest.mark.parametrize('assume_units', [None, 'mbar'])
@pytest.mark.parametrize('method', ['natural_neighbor', 'cressman', 'barnes', 'linear',
                                    'nearest', 'rbf', 'cubic'])
def test_interpolate_to_points(method, assume_units, test_data):
    r"""Test main grid interpolation function."""
    xp, yp, z = test_data
    obs_points = np.vstack([xp, yp]).transpose() * 10

    with get_test_data('interpolation_test_points.npz') as fobj:
        test_points = np.load(fobj)['points']

    if method == 'cressman':
        extra_kw = {'search_radius': 200, 'minimum_neighbors': 1}
    elif method == 'barnes':
        extra_kw = {'search_radius': 400, 'minimum_neighbors': 1, 'gamma': 1}
    else:
        extra_kw = {}

    with get_test_data(f'{method}_test.npz') as fobj:
        truth = np.load(fobj)['img'].reshape(-1)

    if assume_units:
        z = units.Quantity(z, assume_units)
        truth = units.Quantity(truth, assume_units)

    img = interpolate_to_points(obs_points, z, test_points, interp_type=method, **extra_kw)
    assert_array_almost_equal(truth, img)



#!/usr/lib/env python
#
# Module Documentation Override Check
#
# Verifies that any modules present in the _templates/overrides directory have all of their
# their exported functions included in the doc file.
import importlib
from pathlib import Path
import sys

modules_to_skip = ['metpy.xarray']


failed = False
for full_path in (Path('_templates') / 'overrides').glob('metpy.*.rst'):

    module = full_path.with_suffix('').name
    if module in modules_to_skip:
        continue

    # Get all functions in the module
    i = importlib.import_module(module)
    functions = set(i.__all__)

    # Get all lines in the file
    with open(full_path) as f:
        lines = f.read().splitlines()
    lines = {line.strip() for line in lines}

    # Check for any missing functions
    missing_functions = functions - lines

    if missing_functions:
        failed = True
        print('ERROR - The following functions are missing from the override file ' +
              str(full_path.name) + ': ' + ', '.join(missing_functions), file=sys.stderr)

# Report status
if failed:
    sys.exit(1)
else:
    print('Override check successful.')



#  Copyright (c) 2019 MetPy Developers.
#  Distributed under the terms of the BSD 3-Clause License.
#  SPDX-License-Identifier: BSD-3-Clause
"""
=================
Combined Plotting
=================

Demonstrate the use of MetPy's simplified plotting interface combining multiple plots.

Also shows how to control the maps that are plotted. Plots sample NARR data.
"""

import xarray as xr

from metpy.cbook import get_test_data
from metpy.plots import ContourPlot, ImagePlot, MapPanel, PanelContainer
from metpy.units import units

# Use sample NARR data for plotting
narr = xr.open_dataset(get_test_data('narr_example.nc', as_file_obj=False))

###########################
# Create a contour plot of temperature
contour = ContourPlot()
contour.data = narr
contour.field = 'Temperature'
contour.level = 850 * units.hPa
contour.linecolor = 'red'
contour.contours = 15

###########################
# Create an image plot of Geopotential height
img = ImagePlot()
img.data = narr
img.field = 'Geopotential_height'
img.level = 850 * units.hPa


###########################
# Plot the data on a map
panel = MapPanel()
panel.area = 'us'
panel.layers = ['coastline', 'borders', 'states', 'rivers', 'ocean', 'land']
panel.title = 'NARR Example'
panel.plots = [contour, img]

pc = PanelContainer()
pc.size = (10, 8)
pc.panels = [panel]
pc.show()



# Copyright (c) 2008,2015 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
r"""Contains calculations related to turbulence and time series perturbations."""

import numpy as np

from .tools import make_take
from ..package_tools import Exporter
from ..xarray import preprocess_and_wrap

exporter = Exporter(globals())


@exporter.export
@preprocess_and_wrap(wrap_like='ts')
def get_perturbation(ts, axis=-1):
    r"""Compute the perturbation from the mean of a time series.

    Parameters
    ----------
    ts : array-like
         The time series from which you wish to find the perturbation
         time series (perturbation from the mean).

    Returns
    -------
    array-like
        The perturbation time series.

    Other Parameters
    ----------------
    axis : int
           The index of the time axis. Default is -1

    Notes
    -----
    The perturbation time series produced by this function is defined as
    the perturbations about the mean:

    .. math:: x(t)^{\prime} = x(t) - \overline{x(t)}

    """
    mean = ts.mean(axis=axis)[make_take(ts.ndim, axis)(None)]
    return ts - mean


@exporter.export
@preprocess_and_wrap(wrap_like='u')
def tke(u, v, w, perturbation=False, axis=-1):
    r"""Compute turbulence kinetic energy.

    Compute the turbulence kinetic energy (e) from the time series of the
    velocity components.

    Parameters
    ----------
    u : array-like
        The wind component along the x-axis
    v : array-like
        The wind component along the y-axis
    w : array-like
        The wind component along the z-axis
    perturbation : bool, optional
                   True if the `u`, `v`, and `w` components of wind speed
                   supplied to the function are perturbation velocities.
                   If False, perturbation velocities will be calculated by
                   removing the mean value from each component.

    Returns
    -------
    array-like
        The corresponding turbulence kinetic energy value

    Other Parameters
    ----------------
    axis : int
           The index of the time axis. Default is -1

    See Also
    --------
    get_perturbation : Used to compute perturbations if `perturbation`
                       is False.

    Notes
    -----
    Turbulence Kinetic Energy is computed as:

    .. math:: e = 0.5 \sqrt{\overline{u^{\prime2}} +
                            \overline{v^{\prime2}} +
                            \overline{w^{\prime2}}},

    where the velocity components

    .. math:: u^{\prime}, v^{\prime}, u^{\prime}

    are perturbation velocities. For more information on the subject, please
    see [Garratt1994]_.

    """
    if not perturbation:
        u = get_perturbation(u, axis=axis)
        v = get_perturbation(v, axis=axis)
        w = get_perturbation(w, axis=axis)

    u_cont = np.mean(u**2, axis=axis)
    v_cont = np.mean(v**2, axis=axis)
    w_cont = np.mean(w**2, axis=axis)

    return 0.5 * np.sqrt(u_cont + v_cont + w_cont)


@exporter.export
@preprocess_and_wrap(wrap_like='vel')
def kinematic_flux(vel, b, perturbation=False, axis=-1):
    r"""Compute the kinematic flux from two time series.

    Compute the kinematic flux from the time series of two variables `vel`
    and b. Note that to be a kinematic flux, at least one variable must be
    a component of velocity.

    Parameters
    ----------
    vel : array-like
        A component of velocity

    b : array-like
        May be a component of velocity or a scalar variable (e.g. Temperature)

    perturbation : bool, optional
        `True` if the `vel` and `b` variables are perturbations. If `False`, perturbations
        will be calculated by removing the mean value from each variable. Defaults to `False`.

    Returns
    -------
    array-like
        The corresponding kinematic flux

    Other Parameters
    ----------------
    axis : int, optional
           The index of the time axis, along which the calculations will be
           performed. Defaults to -1

    Notes
    -----
    A kinematic flux is computed as

    .. math:: \overline{u^{\prime} s^{\prime}}

    where at the prime notation denotes perturbation variables, and at least
    one variable is perturbation velocity. For example, the vertical kinematic
    momentum flux (two velocity components):

    .. math:: \overline{u^{\prime} w^{\prime}}

    or the vertical kinematic heat flux (one velocity component, and one
    scalar):

    .. math:: \overline{w^{\prime} T^{\prime}}

    If perturbation variables are passed into this function (i.e.
    `perturbation` is True), the kinematic flux is computed using the equation
    above.

    However, the equation above can be rewritten as

    .. math:: \overline{us} - \overline{u}~\overline{s}

    which is computationally more efficient. This is how the kinematic flux
    is computed in this function if `perturbation` is False.

    For more information on the subject, please see [Garratt1994]_.

    """
    kf = np.mean(vel * b, axis=axis)
    if not perturbation:
        kf -= np.mean(vel, axis=axis) * np.mean(b, axis=axis)
    return np.atleast_1d(kf)


@exporter.export
@preprocess_and_wrap(wrap_like='u')
def friction_velocity(u, w, v=None, perturbation=False, axis=-1):
    r"""Compute the friction velocity from the time series of velocity components.

    Compute the friction velocity from the time series of the x, z,
    and optionally y, velocity components.

    Parameters
    ----------
    u : array-like
        The wind component along the x-axis
    w : array-like
        The wind component along the z-axis
    v : array-like, optional
        The wind component along the y-axis.

    perturbation : bool, optional
                   True if the `u`, `w`, and `v` components of wind speed
                   supplied to the function are perturbation velocities.
                   If False, perturbation velocities will be calculated by
                   removing the mean value from each component.

    Returns
    -------
    array-like
        The corresponding friction velocity

    Other Parameters
    ----------------
    axis : int
           The index of the time axis. Default is -1

    See Also
    --------
    kinematic_flux : Used to compute the x-component and y-component
                     vertical kinematic momentum flux(es) used in the
                     computation of the friction velocity.

    Notes
    -----
    The Friction Velocity is computed as:

    .. math:: u_{*} = \sqrt[4]{\left(\overline{u^{\prime}w^{\prime}}\right)^2 +
                               \left(\overline{v^{\prime}w^{\prime}}\right)^2},

    where :math: \overline{u^{\prime}w^{\prime}} and
    :math: \overline{v^{\prime}w^{\prime}}
    are the x-component and y-components of the vertical kinematic momentum
    flux, respectively. If the optional v component of velocity is not
    supplied to the function, the computation of the friction velocity is
    reduced to

    .. math:: u_{*} = \sqrt[4]{\left(\overline{u^{\prime}w^{\prime}}\right)^2}

    For more information on the subject, please see [Garratt1994]_.

    """
    uw = kinematic_flux(u, w, perturbation=perturbation, axis=axis)
    kf = uw**2
    if v is not None:
        vw = kinematic_flux(v, w, perturbation=perturbation, axis=axis)
        kf += vw**2
    # the friction velocity is the 4th root of the kinematic momentum flux
    # As an optimization, first do inplace square root, then return the
    # square root of that. This is faster than np.power(..., 0.25)
    np.sqrt(kf, out=kf)
    return np.sqrt(kf)



# Copyright (c) 2008,2015,2016,2018,2021 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Constant and thermophysical property values expressed as quantities."""

from ..package_tools import Exporter
from ..units import units

exporter = Exporter(globals())

# Export all the variables defined in this block
with exporter:
    # Earth
    earth_gravity = g = units.Quantity(9.80665, 'm / s^2')
    Re = earth_avg_radius = units.Quantity(6371008.7714, 'm')
    G = gravitational_constant = units.Quantity(6.67430e-11, 'm^3 / kg / s^2')
    GM = geocentric_gravitational_constant = units.Quantity(3986005e8, 'm^3 / s^2')
    omega = earth_avg_angular_vel = units.Quantity(7292115e-11, 'rad / s')
    d = earth_sfc_avg_dist_sun = units.Quantity(149597870700., 'm')
    S = earth_solar_irradiance = units.Quantity(1360.8, 'W / m^2')
    delta = earth_max_declination = units.Quantity(23.45, 'degrees')
    earth_orbit_eccentricity = units.Quantity(0.0167, 'dimensionless')
    earth_mass = me = geocentric_gravitational_constant / gravitational_constant

    # molar gas constant
    R = units.Quantity(8.314462618, 'J / mol / K')

    # Water
    Mw = water_molecular_weight = units.Quantity(18.015268, 'g / mol')
    Rv = water_gas_constant = (R / Mw).to('J / kg / K')
    rho_l = density_water = units.Quantity(999.97495, 'kg / m^3')
    wv_specific_heat_ratio = units.Quantity(1.330, 'dimensionless')
    Cp_v = wv_specific_heat_press = (
        wv_specific_heat_ratio * Rv / (wv_specific_heat_ratio - 1)
    )
    Cv_v = wv_specific_heat_vol = Cp_v / wv_specific_heat_ratio
    Cp_l = water_specific_heat = units.Quantity(4.2194, 'kJ / kg / K').to('J / kg / K')
    Lv = water_heat_vaporization = units.Quantity(2.50084e6, 'J / kg')
    Lf = water_heat_fusion = units.Quantity(3.337e5, 'J / kg')
    Cp_i = ice_specific_heat = units.Quantity(2090, 'J / kg / K')
    rho_i = density_ice = units.Quantity(917, 'kg / m^3')
    sat_pressure_0c = units.Quantity(6.112, 'millibar')

    # Dry air
    Md = dry_air_molecular_weight = units.Quantity(28.96546e-3, 'kg / mol')
    Rd = dry_air_gas_constant = R / Md
    dry_air_spec_heat_ratio = units.Quantity(1.4, 'dimensionless')
    Cp_d = dry_air_spec_heat_press = (
        dry_air_spec_heat_ratio * Rd / (dry_air_spec_heat_ratio - 1)
    )
    Cv_d = dry_air_spec_heat_vol = Cp_d / dry_air_spec_heat_ratio
    rho_d = dry_air_density_stp = (
        units.Quantity(1000., 'mbar') / (Rd * units.Quantity(273.15, 'K'))
    ).to('kg / m^3')

    # General meteorology constants
    P0 = pot_temp_ref_press = units.Quantity(1000., 'mbar')
    kappa = poisson_exponent = (Rd / Cp_d).to('dimensionless')
    gamma_d = dry_adiabatic_lapse_rate = (g / Cp_d).to('K / km')
    epsilon = molecular_weight_ratio = (Mw / Md).to('dimensionless')

del Exporter



# Copyright (c) 2014,2015 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

# flake8: noqa
# Generated file -- do not modify
descriptions = {"T1_Output_Frames": "The number of octets received on interface, including frame octets (octet)",
                "T1_Input_Frames": "The number of octets sent on interface, including frame octets (octet)",
                "Router_Memory_Used": "Bytes currently in use by applications on managed device (byte)",
                "Router_Memory_Free": "Bytes currently free on managed device (byte)",
                "Router_Memory_Utilization": "%",
                "CSU_Loss_of_Signal": "Number of times Loss of Signal event detected",
                "CSU_Loss_of_Frames": "Number of times Loss of Frames event detected",
                "CSU_Yellow_Alarms": "Number of times Resource Availability Indication (RAI) (yellow) alarm received.",
                "CSU_Blue_Alarms": "Number of times Alarm Indication Signal (AIS) (blue) alarm received.",
                "CSU_24hr_Errored_Seconds": "Number of errored seconds in previous 24 hours. (s)",
                "CSU_24hr_Severely_Errored_Seconds": "Number of severely errored seconds in previous 24 hours. (s)",
                "CSU_24hr_Severely_Errored_Framing_Seconds": "Number of severely errored framing seconds in previous 24 hours. (s)",
                "CSU_24hr_Unavailable_Seconds": "Number of unavailable seconds in previous 24 hours. (s)",
                "CSU_24hr_Controlled_Slip_Seconds": "Number of controlled slip seconds in previous 24 hours. (s)",
                "CSU_24hr_Path_Coding_Violations": "Number of path coding violations in previous 24 hours.",
                "CSU_24hr_Line_Errored_Seconds": "Number of line errored seconds in previous 24 hours. (s)",
                "CSU_24hr_Bursty_Errored_Seconds": "Number of bursty errored seconds in previous 24 hours. (s)",
                "CSU_24hr_Degraded_Minutes": "Number of degraded minutes in previous 24 hours. (min)",
                "LAN_Switch_Memory_Used": "Bytes currently in use by applications on this device (byte)",
                "LAN_Switch_Memory_Free": "Bytes currently free on this device (byte)",
                "LAN_Switch_Memory_Utilization": "%",
                "NTP_Rejected_Packets": "Number of packets rejected by NTP application layer",
                "NTP_Estimated_Time_Error": "Current estimated time error of the time server (usec)",
                "GPS_Satellites": "Current number of GPS satellites used in position and time fix calculation",
                "GPS_Max_Signal_Strength": "Strongest signal strength of all tracking satellites as seen by receiver (dB)",
                "IPC_Status": "Status of the communications between channels on a redundant system. N/A on a Single channel system.",
                "Commanded_Channel_Control": "Indicates which channel the RDA has commanded to be the controlling channel. This is not necessarily the channel which is in control.",
                "DAU_Test_0": "Tests the performance of the DAU A/D Mutiplexer with a known low voltage input.",
                "DAU_Test_1": "Tests the performance of the DAU A/D Mutiplexer with a known medium voltage input.",
                "DAU_Test_2": "Tests the performance of the DAU A/D Mutiplexer with a known high voltage input.",
                "AME_Internal_Temperature": "deg C",
                "AME_Receiver_Module_Temperature": "deg C",
                "AME_BITE_CAL_Module_Temperature": "deg C",
                "AME_Peltier_Pulse_Width_Modulation": "%",
                "AME_pos_3_3V_PS_Voltage": "V",
                "AME_pos_5V_PS_Voltage": "V",
                "AME_pos_6_5V_PS_Voltage": "V",
                "AME_pos_15V_PS_Voltage": "V",
                "AME_pos_48V_PS_Voltage": "V",
                "AME_STALO_Power": "V",
                "Peltier_Current": "A",
                "ADC_Calibration_Reference_Voltage": "V",
                "AME_Peltier_Inside_Fan_Current": "A",
                "AME_Peltier_Outside_Fan_Current": "A",
                "Horizontal_TR_Limiter_Voltage": "V",
                "Vertical_TR_Limiter_Voltage": "V",
                "ADC_Calibration_Offset_Voltage": "mV",
                "UPS_Time_on_Battery": "s",
                "UPS_Battery_Temperature": "deg C",
                "UPS_Output_Voltage": "V",
                "UPS_Output_Frequency": "Hz",
                "UPS_Output_Current": "A",
                "Power_Administrator_Load": "A",
                "Transmitter_RF_Power__Sensor": "mW",
                "Horizontal_XMTR_Peak_Power": "kW",
                "XMTR_Peak_Power": "kW",
                "Vertical_XMTR_Peak_Power": "kW",
                "XMTR_RF_Avg_Power": "W",
                "Receiver_Bias": "dB",
                "Transmit_Imbalance": "dB",
                "Equipment_Shelter_Temperature": "deg C",
                "Outside_Ambient_Temperature": "deg C",
                "Transmitter_Leaving_Air_Temp": "deg C",
                "AC_Unit__1_Discharge_Air_Temp": "deg C",
                "Generator_Shelter_Temperature": "deg C",
                "Radome_Air_Temperature": "deg C",
                "AC_Unit__2_Discharge_Air_Temp": "deg C",
                "DAU_pos_15v_PS": "V",
                "DAU_neg_15v_PS": "V",
                "DAU_pos_28v_PS": "V",
                "DAU_pos_5v_PS": "V",
                "Converted_Generator_Fuel_Level": "%",
                "pos_28v_PS": "V",
                "Pedestal_pos_15v_PS": "V",
                "Encoder_pos_5v_PS": "V",
                "Pedestal_pos_5v_PS": "V",
                "Pedestal_neg_15v_PS": "V",
                "Horizontal_Short_Pulse_Noise": "dBm",
                "Horizontal_Long_Pulse_Noise": "dBm",
                "Horizontal_Noise_Temperature": "K",
                "Vertical_Noise_Long_Pulse": "dBm",
                "Vertical_Noise_Temperature": "K",
                "Horizontal_Dynamic_Range": "dB",
                "Horizontal_Delta_dBZ0": "dB",
                "Vertical_Delta_dBZ0": "dB",
                "KD_Peak_Measured": "dBm",
                "Short_Pulse__Horizontal_dBZ0": "dBZ",
                "Long_Pulse__Horizontal_dBZ0": "dBZ",
                "Horizontal_I0": "dBm",
                "Vertical_I0": "dBm",
                "Vertical_Dynamic_Range": "dB",
                "Short_Pulse__Vertical_dBZ0": "dBZ",
                "Long_Pulse__Vertical_dBZ0": "dBZ",
                "Horizontal_Power_Sense": "dBm",
                "Vertical_Power_Sense": "dBm",
                "ZDR_Bias": "dB",
                "Clutter_Suppression_Delta": "dB",
                "Clutter_Suppression_Unfiltered_Power": "dBZ",
                "Clutter_Suppression_Filtered_Power": "dBZ",
                "Transmit_Burst_Power": "dBm",
                "Transmit_Burst_Phase": "deg"}

fields = [(None, "2x"),
          ("Loop_Back_Test_Status", "H"),
          ("T1_Output_Frames", "L"),
          ("T1_Input_Frames", "L"),
          ("Router_Memory_Used", "L"),
          ("Router_Memory_Free", "L"),
          ("Router_Memory_Utilization", "H"),
          (None, "2x"),
          ("CSU_Loss_of_Signal", "L"),
          ("CSU_Loss_of_Frames", "L"),
          ("CSU_Yellow_Alarms", "L"),
          ("CSU_Blue_Alarms", "L"),
          ("CSU_24hr_Errored_Seconds", "L"),
          ("CSU_24hr_Severely_Errored_Seconds", "L"),
          ("CSU_24hr_Severely_Errored_Framing_Seconds", "L"),
          ("CSU_24hr_Unavailable_Seconds", "L"),
          ("CSU_24hr_Controlled_Slip_Seconds", "L"),
          ("CSU_24hr_Path_Coding_Violations", "L"),
          ("CSU_24hr_Line_Errored_Seconds", "L"),
          ("CSU_24hr_Bursty_Errored_Seconds", "L"),
          ("CSU_24hr_Degraded_Minutes", "L"),
          ("LAN_Switch_Memory_Used", "L"),
          ("LAN_Switch_Memory_Free", "L"),
          ("LAN_Switch_Memory_Utilization", "H"),
          (None, "2x"),
          ("NTP_Rejected_Packets", "L"),
          ("NTP_Estimated_Time_Error", "l"),
          ("GPS_Satellites", "l"),
          ("GPS_Max_Signal_Strength", "l"),
          ("IPC_Status", "H"),
          ("Commanded_Channel_Control", "H"),
          ("DAU_Test_0", "H"),
          ("DAU_Test_1", "H"),
          ("DAU_Test_2", "H"),
          ("AME_Polarization", "H"),
          ("AME_Internal_Temperature", "f"),
          ("AME_Receiver_Module_Temperature", "f"),
          ("AME_BITE_CAL_Module_Temperature", "f"),
          ("AME_Peltier_Pulse_Width_Modulation", "H"),
          ("AME_Peltier_Status", "H"),
          ("AME_A_D_Converter_Status", "H"),
          ("AME_State", "H"),
          ("AME_pos_3_3V_PS_Voltage", "f"),
          ("AME_pos_5V_PS_Voltage", "f"),
          ("AME_pos_6_5V_PS_Voltage", "f"),
          ("AME_pos_15V_PS_Voltage", "f"),
          ("AME_pos_48V_PS_Voltage", "f"),
          ("AME_STALO_Power", "f"),
          ("Peltier_Current", "f"),
          ("ADC_Calibration_Reference_Voltage", "f"),
          ("AME_Mode", "H"),
          ("AME_Peltier_Mode", "H"),
          ("AME_Peltier_Inside_Fan_Current", "f"),
          ("AME_Peltier_Outside_Fan_Current", "f"),
          ("Horizontal_TR_Limiter_Voltage", "f"),
          ("Vertical_TR_Limiter_Voltage", "f"),
          ("ADC_Calibration_Offset_Voltage", "f"),
          ("ADC_Calibration_Gain_Correction", "f"),
          ("Power_UPS_Battery_Status", "L"),
          ("UPS_Time_on_Battery", "L"),
          ("UPS_Battery_Temperature", "f"),
          ("UPS_Output_Voltage", "f"),
          ("UPS_Output_Frequency", "f"),
          ("UPS_Output_Current", "f"),
          ("Power_Administrator_Load", "f"),
          (None, "48x"),
          ("pos_5_VDC_PS", "H"),
          ("pos_15_VDC_PS", "H"),
          ("pos_28_VDC_PS", "H"),
          ("neg_15_VDC_PS", "H"),
          ("pos_45_VDC_PS", "H"),
          ("Filament_PS_Voltage", "H"),
          ("Vacuum_Pump_PS_Voltage", "H"),
          ("Focus_Coil_PS_Voltage", "H"),
          ("Filament_PS", "H"),
          ("Klystron_Warmup", "H"),
          ("Transmitter_Available", "H"),
          ("WG_Switch_Position", "H"),
          ("WG_PFN_Transfer_Interlock", "H"),
          ("Maintenance_Mode", "H"),
          ("Maintenance_Required", "H"),
          ("PFN_Switch_Position", "H"),
          ("Modulator_Overload", "H"),
          ("Modulator_Inv_Current", "H"),
          ("Modulator_Switch_Fail", "H"),
          ("Main_Power_Voltage", "H"),
          ("Charging_System_Fail", "H"),
          ("Inverse_Diode_Current", "H"),
          ("Trigger_Amplifier", "H"),
          ("Circulator_Temperature", "H"),
          ("Spectrum_Filter_Pressure", "H"),
          ("WG_ARC_VSWR", "H"),
          ("Cabinet_Interlock", "H"),
          ("Cabinet_Air_Temperature", "H"),
          ("Cabinet_Airflow", "H"),
          ("Klystron_Current", "H"),
          ("Klystron_Filament_Current", "H"),
          ("Klystron_Vacion_Current", "H"),
          ("Klystron_Air_Temperature", "H"),
          ("Klystron_Airflow", "H"),
          ("Modulator_Switch_Maintenance", "H"),
          ("Post_Charge_Regulator_Maintenance", "H"),
          ("WG_Pressure_Humidity", "H"),
          ("Transmitter_Overvoltage", "H"),
          ("Transmitter_Overcurrent", "H"),
          ("Focus_Coil_Current", "H"),
          ("Focus_Coil_Airflow", "H"),
          ("Oil_Temperature", "H"),
          ("PRF_Limit", "H"),
          ("Transmitter_Oil_Level", "H"),
          ("Transmitter_Battery_Charging", "H"),
          ("High_Voltage__HV__Status", "H"),
          ("Transmitter_Recycling_Summary", "H"),
          ("Transmitter_Inoperable", "H"),
          ("Transmitter_Air_Filter", "H"),
          ("Zero_Test_Bit_0", "H"),
          ("Zero_Test_Bit_1", "H"),
          ("Zero_Test_Bit_2", "H"),
          ("Zero_Test_Bit_3", "H"),
          ("Zero_Test_Bit_4", "H"),
          ("Zero_Test_Bit_5", "H"),
          ("Zero_Test_Bit_6", "H"),
          ("Zero_Test_Bit_7", "H"),
          ("One_Test_Bit_0", "H"),
          ("One_Test_Bit_1", "H"),
          ("One_Test_Bit_2", "H"),
          ("One_Test_Bit_3", "H"),
          ("One_Test_Bit_4", "H"),
          ("One_Test_Bit_5", "H"),
          ("One_Test_Bit_6", "H"),
          ("One_Test_Bit_7", "H"),
          ("XMTR_DAU_Interface", "H"),
          ("Transmitter_Summary_Status", "H"),
          (None, "2x"),
          ("Transmitter_RF_Power__Sensor", "f"),
          ("Horizontal_XMTR_Peak_Power", "f"),
          ("XMTR_Peak_Power", "f"),
          ("Vertical_XMTR_Peak_Power", "f"),
          ("XMTR_RF_Avg_Power", "f"),
          ("XMTR_Power_Meter_Zero", "H"),
          (None, "2x"),
          ("XMTR_Recycle_Count", "L"),
          ("Receiver_Bias", "f"),
          ("Transmit_Imbalance", "f"),
          (None, "12x"),
          ("AC_Unit__1_Compressor_Shut_off", "H"),
          ("AC_Unit__2_Compressor_Shut_off", "H"),
          ("Generator_Maintenance_Required", "H"),
          ("Generator_Battery_Voltage", "H"),
          ("Generator_Engine", "H"),
          ("Generator_Volt_Frequency", "H"),
          ("Power_Source", "H"),
          ("Transitional_Power_Source__TPS", "H"),
          ("Generator_Auto_Run_Off_Switch", "H"),
          ("Aircraft_Hazard_Lighting", "H"),
          ("DAU_UART", "H"),
          (None, "20x"),
          ("Equipment_Shelter_Fire_Detection_System", "H"),
          ("Equipment_Shelter_Fire_Smoke", "H"),
          ("Generator_Shelter_Fire_Smoke", "H"),
          ("Utility_Voltage_Frequency", "H"),
          ("Site_Security_Alarm", "H"),
          ("Security_Equipment", "H"),
          ("Security_System", "H"),
          ("Receiver_Connected_to_Antenna", "H"),
          ("Radome_Hatch", "H"),
          ("AC_Unit__1_Filter_Dirty", "H"),
          ("AC_Unit__2_Filter_Dirty", "H"),
          ("Equipment_Shelter_Temperature", "f"),
          ("Outside_Ambient_Temperature", "f"),
          ("Transmitter_Leaving_Air_Temp", "f"),
          ("AC_Unit__1_Discharge_Air_Temp", "f"),
          ("Generator_Shelter_Temperature", "f"),
          ("Radome_Air_Temperature", "f"),
          ("AC_Unit__2_Discharge_Air_Temp", "f"),
          ("DAU_pos_15v_PS", "f"),
          ("DAU_neg_15v_PS", "f"),
          ("DAU_pos_28v_PS", "f"),
          ("DAU_pos_5v_PS", "f"),
          ("Converted_Generator_Fuel_Level", "H"),
          (None, "14x"),
          ("pos_28v_PS", "f"),
          ("Pedestal_pos_15v_PS", "f"),
          ("Encoder_pos_5v_PS", "f"),
          ("Pedestal_pos_5v_PS", "f"),
          ("Pedestal_neg_15v_PS", "f"),
          ("pos_150V_Overvoltage", "H"),
          ("pos_150V_Undervoltage", "H"),
          ("Elevation_Servo_Amp_Inhibit", "H"),
          ("Elevation_Servo_Amp_Short_Circuit", "H"),
          ("Elevation_Servo_Amp_Overtemp", "H"),
          ("Elevation_Motor_Overtemp", "H"),
          ("Elevation_Stow_Pin", "H"),
          ("Elevation_PCU_Parity", "H"),
          ("Elevation_Dead_Limit", "H"),
          ("Elevation_pos_Normal_Limit", "H"),
          ("Elevation_neg_Normal_Limit", "H"),
          ("Elevation_Encoder_Light", "H"),
          ("Elevation_Gearbox_Oil", "H"),
          ("Elevation_Handwheel", "H"),
          ("Elevation_Amp_PS", "H"),
          ("Azimuth_Servo_Amp_Inhibit", "H"),
          ("Azimuth_Servo_Amp_Short_Circuit", "H"),
          ("Azimuth_Servo_Amp_Overtemp", "H"),
          ("Azimuth_Motor_Overtemp", "H"),
          ("Azimuth_Stow_Pin", "H"),
          ("Azimuth_PCU_Parity", "H"),
          ("Azimuth_Encoder_Light", "H"),
          ("Azimuth_Gearbox_Oil", "H"),
          ("Azimuth_Bull_Gear_Oil", "H"),
          ("Azimuth_Handwheel", "H"),
          ("Azimuth_Servo_Amp_PS", "H"),
          ("Servo", "H"),
          ("Pedestal_Interlock_Switch", "H"),
          (None, "2x"),
          (None, "2x"),
          ("Self_Test_1_Status", "H"),
          ("Self_Test_2_Status", "H"),
          ("Self_Test_2_Data", "H"),
          (None, "14x"),
          ("COHO_Clock", "H"),
          ("Rf_Generator_Frequency_Select_Oscillator", "H"),
          ("Rf_Generator_RF_STALO", "H"),
          ("Rf_Generator_Phase_Shifted_COHO", "H"),
          ("pos_9v_Receiver_PS", "H"),
          ("pos_5v_Receiver_PS", "H"),
          ("O_18v_Receiver_PS", "H"),
          ("neg_9v_Receiver_PS", "H"),
          ("pos_5v_Single_Channel_RDAIU_PS", "H"),
          (None, "2x"),
          ("Horizontal_Short_Pulse_Noise", "f"),
          ("Horizontal_Long_Pulse_Noise", "f"),
          ("Horizontal_Noise_Temperature", "f"),
          (None, "4x"),
          ("Vertical_Noise_Long_Pulse", "f"),
          ("Vertical_Noise_Temperature", "f"),
          ("Horizontal_Linearity", "f"),
          ("Horizontal_Dynamic_Range", "f"),
          ("Horizontal_Delta_dBZ0", "f"),
          ("Vertical_Delta_dBZ0", "f"),
          ("KD_Peak_Measured", "f"),
          (None, "4x"),
          ("Short_Pulse__Horizontal_dBZ0", "f"),
          ("Long_Pulse__Horizontal_dBZ0", "f"),
          ("Velocity__Processed", "H"),
          ("Width__Processed", "H"),
          ("Velocity__RF_Gen", "H"),
          ("Width__RF_Gen", "H"),
          ("Horizontal_I0", "f"),
          ("Vertical_I0", "f"),
          ("Vertical_Dynamic_Range", "f"),
          ("Short_Pulse__Vertical_dBZ0", "f"),
          ("Long_Pulse__Vertical_dBZ0", "f"),
          (None, "4x"),
          (None, "4x"),
          ("Horizontal_Power_Sense", "f"),
          ("Vertical_Power_Sense", "f"),
          ("ZDR_Bias", "f"),
          (None, "12x"),
          ("Clutter_Suppression_Delta", "f"),
          ("Clutter_Suppression_Unfiltered_Power", "f"),
          ("Clutter_Suppression_Filtered_Power", "f"),
          ("Transmit_Burst_Power", "f"),
          ("Transmit_Burst_Phase", "f"),
          (None, "8x"),
          (None, "4x"),
          ("Vertical_Linearity", "f"),
          (None, "8x"),
          ("State_File_Read_Status", "H"),
          ("State_File_Write_Status", "H"),
          ("Bypass_Map_File_Read_Status", "H"),
          ("Bypass_Map_File_Write_Status", "H"),
          (None, "2x"),
          (None, "2x"),
          ("Current_Adaptation_File_Read_Status", "H"),
          ("Current_Adaptation_File_Write_Status", "H"),
          ("Censor_Zone_File_Read_Status", "H"),
          ("Censor_Zone_File_Write_Status", "H"),
          ("Remote_VCP_File_Read_Status", "H"),
          ("Remote_VCP_File_Write_Status", "H"),
          ("Baseline_Adaptation_File_Read_Status", "H"),
          (None, "2x"),
          ("Clutter_Filter_Map_File_Read_Status", "H"),
          ("Clutter_Filter_Map_File_Write_Status", "H"),
          ("General_Disk_I_O_Error", "H"),
          (None, "26x"),
          ("DAU_Comm_Status", "H"),
          ("HCI_Comm_Status", "H"),
          ("Pedestal_Comm_Status", "H"),
          ("Signal_Processor_Comm_Status", "H"),
          ("AME_Communication_Status", "H"),
          ("RMS_Link_Status", "H"),
          ("RPG_Link_Status", "H"),
          (None, "2x"),
          ("Performance_Check_Time", "L"),
          (None, "20x")]



# Copyright (c) 2014,2015 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

# flake8: noqa
# Generated file -- do not modify
descriptions = {"ADAP_FILE_NAME": "NAME OF ADAPTATION DATA FILE",
                "ADAP_FORMAT": "FORMAT OF ADAPTATION DATA FILE",
                "ADAP_REVISION": "REVISION NUMBER OF ADAPTATION DATA FILE",
                "ADAP_DATE": "LAST MODIFIED DATE ADAPTATION DATA FILE",
                "ADAP_TIME": "LAST MODIFIED TIME OF ADAPTATION DATA FILE",
                "K1": "AZIMUTH POSITION GAIN FACTOR (K1)",
                "AZ_LAT": "LATENCY OF DCU AZIMUTH MEASUREMENT (s)",
                "K3": "ELEVATION POSITION GAIN FACTOR (K3)",
                "EL_LAT": "LATENCY OF DCU ELEVATION MEASUREMENT (s)",
                "PARKAZ": "PEDESTAL PARK POSITION IN AZIMUTH (deg)",
                "PARKEL": "PEDESTAL PARK POSITION IN ELEVATION (deg)",
                "A_FUEL_CONV_0": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (0% HGT) (%)",
                "A_FUEL_CONV_1": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (10% HGT) (%)",
                "A_FUEL_CONV_2": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (20% HGT) (%)",
                "A_FUEL_CONV_3": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (30% HGT) (%)",
                "A_FUEL_CONV_4": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (40% HGT) (%)",
                "A_FUEL_CONV_5": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (50% HGT) (%)",
                "A_FUEL_CONV_6": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (60% HGT) (%)",
                "A_FUEL_CONV_7": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (70% HGT) (%)",
                "A_FUEL_CONV_8": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (80% HGT) (%)",
                "A_FUEL_CONV_9": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (90% HGT) (%)",
                "A_FUEL_CONV_10": "GENERATOR FUEL LEVEL HEIGHT/CAPACITY CONVERSION (100% HGT) (%)",
                "A_MIN_SHELTER_TEMP": "MINIMUM EQUIPMENT SHELTER ALARM TEMPERATURE (deg C)",
                "A_MAX_SHELTER_TEMP": "MAXIMUM EQUIPMENT SHELTER ALARM TEMPERATURE (deg C)",
                "A_MIN_SHELTER_AC_TEMP_DIFF": "MINIMUM A/C DISCHARGE AIR TEMPERATURE DIFFERENTIAL (deg C)",
                "A_MAX_XMTR_AIR_TEMP": "MAXIMUM TRANSMITTER LEAVING AIR ALARM TEMPERATURE (deg C)",
                "A_MAX_RAD_TEMP": "MAXIMUM RADOME ALARM TEMPERATURE (deg C)",
                "A_MAX_RAD_TEMP_RISE": "MAXIMUM RADOME MINUS AMBIENT TEMPERATURE DIFFERENCE (deg C)",
                "PED_28V_REG_LIM": "PEDESTAL +28 VOLT POWER SUPPLY TOLERANCE (%)",
                "PED_5V_REG_LIM": "PEDESTAL +5 VOLT POWER SUPPLY TOLERANCE (%)",
                "PED_15V_REG_LIM": "PEDESTAL +/- 15 VOLT POWER SUPPLY TOLERANCE (%)",
                "A_MIN_GEN_ROOM_TEMP": "MINIMUM GENERATOR SHELTER ALARM TEMPERATURE (deg C)",
                "A_MAX_GEN_ROOM_TEMP": "MAXIMUM GENERATOR SHELTER ALARM TEMPERATURE (deg C)",
                "DAU_5V_REG_LIM": "DAU +5 VOLT POWER SUPPLY TOLERANCE (%)",
                "DAU_15V_REG_LIM": "DAU +/- 15 VOLT POWER SUPPLY TOLERANCE (%)",
                "DAU_28V_REG_LIM": "DAU +28 VOLT POWER (%)",
                "EN_5V_REG_LIM": "ENCODER +5 VOLT POWER SUPPLY TOLERANCE (%)",
                "EN_5V_NOM_VOLTS": "ENCODER +5 VOLT POWER SUPPLY NOMINAL VOLTAGE (V)",
                "RPG_CO_LOCATED": "RPG CO-LOCATED",
                "SPEC_FILTER_INSTALLED": "TRANSMITTER SPECTRUM FILTER INSTALLED",
                "TPS_INSTALLED": "TRANSITION POWER SOURCE INSTALLED",
                "RMS_INSTALLED": "FAA RMS INSTALLED",
                "A_HVDL_TST_INT": "PERFORMANCE TEST INTERVAL (h)",
                "A_RPG_LT_INT": "RPG LOOP TEST INTERVAL (min)",
                "A_MIN_STAB_UTIL_PWR_TIME": "REQUIRED INTERVAL TIME FOR STABLE UTILITY POWER (min)",
                "A_GEN_AUTO_EXER_INTERVAL": "MAXIMUM GENERATOR AUTOMATIC EXERCISE INTERVAL (h)",
                "A_UTIL_PWR_SW_REQ_INTERVAL": "RECOMMENDED SWITCH TO UTILITY POWER TIME INTERVAL (min)",
                "A_LOW_FUEL_LEVEL": "LOW FUEL TANK WARNING LEVEL (%)",
                "CONFIG_CHAN_NUMBER": "CONFIGURATION CHANNEL NUMBER",
                "A_RPG_LINK_TYPE": "RPG WIDEBAND LINK TYPE (0 = DIRECT, 1 = MICROWAVE, 2 = FIBER OPTIC)",
                "REDUNDANT_CHAN_CONFIG": "REDUNDANT CHANNEL CONFIGURATION (1 = SINGLE CHAN, 2 = FAA, 3 = NWS REDUNDANT)",
                "ATTEN_TABLE_0": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (0dB) (dB)",
                "ATTEN_TABLE_1": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (1dB) (dB)",
                "ATTEN_TABLE_2": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (2dB) (dB)",
                "ATTEN_TABLE_3": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (3dB) (dB)",
                "ATTEN_TABLE_4": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (4dB) (dB)",
                "ATTEN_TABLE_5": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (5dB) (dB)",
                "ATTEN_TABLE_6": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (6dB) (dB)",
                "ATTEN_TABLE_7": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (7dB) (dB)",
                "ATTEN_TABLE_8": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (8dB) (dB)",
                "ATTEN_TABLE_9": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (9dB) (dB)",
                "ATTEN_TABLE_10": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (10dB) (dB)",
                "ATTEN_TABLE_11": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (11dB) (dB)",
                "ATTEN_TABLE_12": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (12dB) (dB)",
                "ATTEN_TABLE_13": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (13dB) (dB)",
                "ATTEN_TABLE_14": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (14dB) (dB)",
                "ATTEN_TABLE_15": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (15dB) (dB)",
                "ATTEN_TABLE_16": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (16dB) (dB)",
                "ATTEN_TABLE_17": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (17dB) (dB)",
                "ATTEN_TABLE_18": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (18dB) (dB)",
                "ATTEN_TABLE_19": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (19dB) (dB)",
                "ATTEN_TABLE_20": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (20dB) (dB)",
                "ATTEN_TABLE_21": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (21dB) (dB)",
                "ATTEN_TABLE_22": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (22dB) (dB)",
                "ATTEN_TABLE_23": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (23dB) (dB)",
                "ATTEN_TABLE_24": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (24dB) (dB)",
                "ATTEN_TABLE_25": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (25dB) (dB)",
                "ATTEN_TABLE_26": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (26dB) (dB)",
                "ATTEN_TABLE_27": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (27dB) (dB)",
                "ATTEN_TABLE_28": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (28dB) (dB)",
                "ATTEN_TABLE_29": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (29dB) (dB)",
                "ATTEN_TABLE_30": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (30dB) (dB)",
                "ATTEN_TABLE_31": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (31dB) (dB)",
                "ATTEN_TABLE_32": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (32dB) (dB)",
                "ATTEN_TABLE_33": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (33dB) (dB)",
                "ATTEN_TABLE_34": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (34dB) (dB)",
                "ATTEN_TABLE_35": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (35dB) (dB)",
                "ATTEN_TABLE_36": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (36dB) (dB)",
                "ATTEN_TABLE_37": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (37dB) (dB)",
                "ATTEN_TABLE_38": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (38dB) (dB)",
                "ATTEN_TABLE_39": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (39dB) (dB)",
                "ATTEN_TABLE_40": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (40dB) (dB)",
                "ATTEN_TABLE_41": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (41dB) (dB)",
                "ATTEN_TABLE_42": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (42dB) (dB)",
                "ATTEN_TABLE_43": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (43dB) (dB)",
                "ATTEN_TABLE_44": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (44dB) (dB)",
                "ATTEN_TABLE_45": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (45dB) (dB)",
                "ATTEN_TABLE_46": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (46dB) (dB)",
                "ATTEN_TABLE_47": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (47dB) (dB)",
                "ATTEN_TABLE_48": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (48dB) (dB)",
                "ATTEN_TABLE_49": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (49dB) (dB)",
                "ATTEN_TABLE_50": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (50dB) (dB)",
                "ATTEN_TABLE_51": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (51dB) (dB)",
                "ATTEN_TABLE_52": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (52dB) (dB)",
                "ATTEN_TABLE_53": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (53dB) (dB)",
                "ATTEN_TABLE_54": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (54dB) (dB)",
                "ATTEN_TABLE_55": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (55dB) (dB)",
                "ATTEN_TABLE_56": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (56dB) (dB)",
                "ATTEN_TABLE_57": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (57dB) (dB)",
                "ATTEN_TABLE_58": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (58dB) (dB)",
                "ATTEN_TABLE_59": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (59dB) (dB)",
                "ATTEN_TABLE_60": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (60dB) (dB)",
                "ATTEN_TABLE_61": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (61dB) (dB)",
                "ATTEN_TABLE_62": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (62dB) (dB)",
                "ATTEN_TABLE_63": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (63dB) (dB)",
                "ATTEN_TABLE_64": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (64dB) (dB)",
                "ATTEN_TABLE_65": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (65dB) (dB)",
                "ATTEN_TABLE_66": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (66dB) (dB)",
                "ATTEN_TABLE_67": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (67dB) (dB)",
                "ATTEN_TABLE_68": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (68dB) (dB)",
                "ATTEN_TABLE_69": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (69dB) (dB)",
                "ATTEN_TABLE_70": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (70dB) (dB)",
                "ATTEN_TABLE_71": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (71dB) (dB)",
                "ATTEN_TABLE_72": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (72dB) (dB)",
                "ATTEN_TABLE_73": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (73dB) (dB)",
                "ATTEN_TABLE_74": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (74dB) (dB)",
                "ATTEN_TABLE_75": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (75dB) (dB)",
                "ATTEN_TABLE_76": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (76dB) (dB)",
                "ATTEN_TABLE_77": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (77dB) (dB)",
                "ATTEN_TABLE_78": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (78dB) (dB)",
                "ATTEN_TABLE_79": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (79dB) (dB)",
                "ATTEN_TABLE_80": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (80dB) (dB)",
                "ATTEN_TABLE_81": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (81dB) (dB)",
                "ATTEN_TABLE_82": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (82dB) (dB)",
                "ATTEN_TABLE_83": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (83dB) (dB)",
                "ATTEN_TABLE_84": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (84dB) (dB)",
                "ATTEN_TABLE_85": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (85dB) (dB)",
                "ATTEN_TABLE_86": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (86dB) (dB)",
                "ATTEN_TABLE_87": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (87dB) (dB)",
                "ATTEN_TABLE_88": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (88dB) (dB)",
                "ATTEN_TABLE_89": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (89dB) (dB)",
                "ATTEN_TABLE_90": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (90dB) (dB)",
                "ATTEN_TABLE_91": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (91dB) (dB)",
                "ATTEN_TABLE_92": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (92dB) (dB)",
                "ATTEN_TABLE_93": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (93dB) (dB)",
                "ATTEN_TABLE_94": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (94dB) (dB)",
                "ATTEN_TABLE_95": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (95dB) (dB)",
                "ATTEN_TABLE_96": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (96dB) (dB)",
                "ATTEN_TABLE_97": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (97dB) (dB)",
                "ATTEN_TABLE_98": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (98dB) (dB)",
                "ATTEN_TABLE_99": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (99dB) (dB)",
                "ATTEN_TABLE_100": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (100dB) (dB)",
                "ATTEN_TABLE_101": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (101dB) (dB)",
                "ATTEN_TABLE_102": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (102dB) (dB)",
                "ATTEN_TABLE_103": "TEST SIGNAL ATTENUATOR INSERTION LOSSES (103dB) (dB)",
                "PATH_LOSSES_7": "PATH LOSS - VERTICAL IF HELIAX TO 4AT16 (dB)",
                "PATH_LOSSES_11": "PATH LOSS - 4AT14 TO RF LOG AMP 4A29J1 (dB)",
                "PATH_LOSSES_13": "PATH LOSS - 2A9A9 RF DELAY LINE (dB)",
                "PATH_LOSSES_14": "PATH LOSS - 4AT13 TO RF LOG AMP 4A29J1 (dB)",
                "PATH_LOSSES_19": "PATH LOSS - 4AT15 TO RF LOG AMP 4A29J1 (dB)",
                "PATH_LOSSES_28": "PATH LOSS - HORIZONTAL IF HELIAX TO 4AT17 (dB)",
                "PATH_LOSSES_30": "PATH LOSS - A5 ELEVATION ROTARY JOINT (dB)",
                "PATH_LOSSES_31": "PATH LOSS - ELEVATION ROTARY JOINT TO ANTENNA (dB)",
                "PATH_LOSSES_32": "PATH LOSS - WG02 HARMONIC FILTER (dB)",
                "PATH_LOSSES_33": "PATH LOSS - WAVEGUIDE KLYSTRON T0 SWITCH (dB)",
                "PATH_LOSSES_34": "PATH LOSS - 2A1A4 WAVEGUIDE CHANNEL AZIMUTH ROTARY JOINT (dB)",
                "PATH_LOSSES_35": "PATH LOSS - WG06 SPECTRUM FILTER (dB)",
                "PATH_LOSSES_36": "PATH LOSS - COAX TRANSMITTER RF DRIVE TO 4AT15 (dB)",
                "PATH_LOSSES_37": "PATH LOSS - WAVEGUIDE SWITCH TO AZIMUTH ROTARY JOINT (dB)",
                "PATH_LOSSES_38": "PATH LOSS - WAVEGUIDE SWITCH (dB)",
                "PATH_LOSSES_39": "PATH LOSS - WG04 CIRCULATOR (dB)",
                "PATH_LOSSES_40": "PATH LOSS - A6 ARC DETECTOR (dB)",
                "PATH_LOSSES_41": "PATH LOSS - 1DC1 TRANSMITTER COUPLER STRAIGHT THRU (dB)",
                "PATH_LOSSES_42": "PATH LOSS - 1DC1 TRANSMITTER COUPLER COUPLING (dB)",
                "PATH_LOSSES_43": "PATH LOSS - A33 PAD (dB)",
                "PATH_LOSSES_44": "PATH LOSS - COAX TRANSMITTER RF SAMPLE TO A33 PAD (dB)",
                "PATH_LOSSES_45": "PATH LOSS - A20J1_4 POWER SPLITTER (dB)",
                "PATH_LOSSES_46": "PATH LOSS - A20J1_3 POWER SPLITTER (dB)",
                "PATH_LOSSES_47": "PATH LOSS - A20J1_2 POWER SPLITTER (dB)",
                "H_COUPLER_CW_LOSS": "RF PALLET HORIZONTAL COUPLER TEST SIGNAL LOSS (dB)",
                "V_COUPLER_XMT_LOSS": "RF PALLET VERTICAL COUPLER TRANSMITTER LOSS (dB)",
                "PATH_LOSSES_50": "PATH LOSS - WAVEGUIDE AZIMUTH JOINT TO ELEVATION JOINT (dB)",
                "PATH_LOSSES_52": "PATH LOSS - 1AT4 TRANSMITTER COUPLER PAD (dB)",
                "V_COUPLER_CW_LOSS": "RF PALLET VERTICAL COUPLER TEST SIGNAL LOSS (dB)",
                "PATH_LOSSES_54": "PATH LOSS - 4AT13 PAD (dB)",
                "PATH_LOSSES_58": "PATH LOSS - 4AT17 ATTENUATOR (dB)",
                "PATH_LOSSES_59": "PATH LOSS - IFD IF ANTI-ALIAS FILTER (dB)",
                "PATH_LOSSES_60": "PATH LOSS - A20J1_5 POWER SPLITTER (dB)",
                "PATH_LOSSES_61": "PATH LOSS - AT5 50dB ATTENUATOR (dB)",
                "PATH_LOSSES_63": "PATH LOSS - A39 RF_IF BURST MIXER (dB)",
                "PATH_LOSSES_64": "PATH LOSS - AR1 BURST IF AMPLIFIER (dB)",
                "PATH_LOSSES_65": "PATH LOSS - IFD BURST ANTIALIAS FILTER (dB)",
                "PATH_LOSSES_66": "PATH LOSS - DC3 J1_3 6dB COUPLER, THROUGH (dB)",
                "PATH_LOSSES_67": "PATH LOSS - DC3 J1_2 6DB COUPLER, COUPLED (dB)",
                "PATH_LOSSES_68": "PATH LOSS - AT2+AT3 26dB COHO ATTENUATOR (dB)",
                "PATH_LOSSES_69": "PATH LOSS - 4DC3J1 TO 4AT14 (dB)",
                "CHAN_CAL_DIFF": "NONCONTROLLING CHANNEL CALIBRATION DIFFERENCE (dB)",
                "LOG_AMP_FACTOR_1": "RF DETECTOR LOG AMPLIFIER SCALE FACTOR FOR CONVERTING RECEIVER TEST DATA (V/dBm)",
                "LOG_AMP_FACTOR_2": "RF DETECTOR LOG AMPLIFIER BIAS FOR CONVERTING RECEIVER TEST DATA (V)",
                "V_TS_CW": "AME VERTICAL TEST SIGNAL POWER (dBm)",
                "RNSCALE_0": "RECEIVER NOISE NORMALIZATION (-1.0 deg to -0.5 deg)",
                "RNSCALE_1": "RECEIVER NOISE NORMALIZATION (-0.5 deg to 0.0 deg)",
                "RNSCALE_2": "RECEIVER NOISE NORMALIZATION (0.0 deg to 0.5 deg)",
                "RNSCALE_3": "RECEIVER NOISE NORMALIZATION (0.5 deg to 1.0 deg)",
                "RNSCALE_4": "RECEIVER NOISE NORMALIZATION (1.0 deg to 1.5 deg)",
                "RNSCALE_5": "RECEIVER NOISE NORMALIZATION (1.5 deg to 2.0 deg)",
                "RNSCALE_6": "RECEIVER NOISE NORMALIZATION (2.0 deg to 2.5 deg)",
                "RNSCALE_7": "RECEIVER NOISE NORMALIZATION (2.5 deg to 3.0 deg)",
                "RNSCALE_8": "RECEIVER NOISE NORMALIZATION (3.0 deg to 3.5 deg)",
                "RNSCALE_9": "RECEIVER NOISE NORMALIZATION (3.5 deg to 4.0 deg)",
                "RNSCALE_10": "RECEIVER NOISE NORMALIZATION (4.0 deg to 4.5 deg)",
                "RNSCALE_11": "RECEIVER NOISE NORMALIZATION (4.5 deg to 5.0 deg)",
                "RNSCALE_12": "RECEIVER NOISE NORMALIZATION (> 5.0 deg)",
                "ATMOS_0": "TWO WAY ATMOSPHERIC LOSS/KM (-1.0 deg to -0.5 deg) (dB/km)",
                "ATMOS_1": "TWO WAY ATMOSPHERIC LOSS/KM (-0.5 deg to 0.0 deg) (dB/km)",
                "ATMOS_2": "TWO WAY ATMOSPHERIC LOSS/KM (0.0 deg to 0.5 deg) (dB/km)",
                "ATMOS_3": "TWO WAY ATMOSPHERIC LOSS/KM (0.5 deg to 1.0 deg) (dB/km)",
                "ATMOS_4": "TWO WAY ATMOSPHERIC LOSS/KM (1.0 deg to 1.5 deg) (dB/km)",
                "ATMOS_5": "TWO WAY ATMOSPHERIC LOSS/KM (1.5 deg to 2.0 deg) (dB/km)",
                "ATMOS_6": "TWO WAY ATMOSPHERIC LOSS/KM (2.0 deg to 2.5 deg) (dB/km)",
                "ATMOS_7": "TWO WAY ATMOSPHERIC LOSS/KM (2.5 deg to 3.0 deg) (dB/km)",
                "ATMOS_8": "TWO WAY ATMOSPHERIC LOSS/KM (3.0 deg to 3.5 deg) (dB/km)",
                "ATMOS_9": "TWO WAY ATMOSPHERIC LOSS/KM (3.5 deg to 4.0 deg) (dB/km)",
                "ATMOS_10": "TWO WAY ATMOSPHERIC LOSS/KM (4.0 deg to 4.5 deg) (dB/km)",
                "ATMOS_11": "TWO WAY ATMOSPHERIC LOSS/KM (4.5 deg to 5.0 deg) (dB/km)",
                "ATMOS_12": "TWO WAY ATMOSPHERIC LOSS/KM (> 5.0 deg) (dB/km)",
                "EL_INDEX_0": "BYPASS MAP GENERATION ELEVATION ANGLE (0) (deg)",
                "EL_INDEX_1": "BYPASS MAP GENERATION ELEVATION ANGLE (1) (deg)",
                "EL_INDEX_2": "BYPASS MAP GENERATION ELEVATION ANGLE (2) (deg)",
                "EL_INDEX_3": "BYPASS MAP GENERATION ELEVATION ANGLE (3) (deg)",
                "EL_INDEX_4": "BYPASS MAP GENERATION ELEVATION ANGLE (4) (deg)",
                "EL_INDEX_5": "BYPASS MAP GENERATION ELEVATION ANGLE (5) (deg)",
                "EL_INDEX_6": "BYPASS MAP GENERATION ELEVATION ANGLE (6) (deg)",
                "EL_INDEX_7": "BYPASS MAP GENERATION ELEVATION ANGLE (7) (deg)",
                "EL_INDEX_8": "BYPASS MAP GENERATION ELEVATION ANGLE (8) (deg)",
                "EL_INDEX_9": "BYPASS MAP GENERATION ELEVATION ANGLE (9) (deg)",
                "EL_INDEX_10": "BYPASS MAP GENERATION ELEVATION ANGLE (10) (deg)",
                "EL_INDEX_11": "BYPASS MAP GENERATION ELEVATION ANGLE (11) (deg)",
                "TFREQ_MHZ": "TRANSMITTER FREQUENCY (MHz)",
                "BASE_DATA_TCN": "POINT CLUTTER SUPPRESSION THRESHOLD (TCN) (dB)",
                "REFL_DATA_TOVER": "RANGE UNFOLDING OVERLAY THRESHOLD (TOVER) (dB)",
                "TAR_H_DBZ0_LP": "HORIZONTAL TARGET SYSTEM CALIBRATION (dBZ0) FOR LONG PULSE (dBZ)",
                "TAR_V_DBZ0_LP": "VERTICAL TARGET SYSTEM CALIBRATION (dBZ0) FOR LONG PULSE (dBZ)",
                "INIT_PHI_DP": "INITIAL SYSTEM DIFFERENTIAL PHASE (deg)",
                "NORM_INIT_PHI_DP": "NORMALIZED INITIAL SYSTEM DIFFERENTIAL PHASE (deg)",
                "LX_LP": "MATCHED FILTER LOSS FOR LONG PULSE (dB)",
                "LX_SP": "MATCHED FILTER LOSS FOR SHORT PULSE (dB)",
                "METEOR_PARAM": "/K/**2 HYDROMETEOR REFRACTIVITY FACTOR",
                "BEAMWIDTH": "ANTENNA BEAMWIDTH (deg)",
                "ANTENNA_GAIN": "ANTENNA GAIN INCLUDING RADOME (dB)",
                "VEL_MAINT_LIMIT": "VELOCITY CHECK DELTA MAINTENANCE LIMIT (m/s)",
                "WTH_MAINT_LIMIT": "SPECTRUM WIDTH CHECK DELTA MAINTENANCE LIMIT (m/s)",
                "VEL_DEGRAD_LIMIT": "VELOCITY CHECK DELTA DEGRADE LIMIT (m/s)",
                "WTH_DEGRAD_LIMIT": "SPECTRUM WIDTH CHECK DELTA DEGRADE LIMIT (m/s)",
                "H_NOISETEMP_DGRAD_LIMIT": "HORIZONTAL SYSTEM NOISE TEMP DEGRADE LIMIT (K)",
                "H_NOISETEMP_MAINT_LIMIT": "HORIZONTAL SYSTEM NOISE TEMP MAINTENANCE LIMIT (K)",
                "V_NOISETEMP_DGRAD_LIMIT": "VERTICAL SYSTEM NOISE TEMP DEGRADE LIMIT (K)",
                "V_NOISETEMP_MAINT_LIMIT": "VERTICAL SYSTEM NOISE TEMP MAINTENANCE LIMIT (K)",
                "KLY_DEGRADE_LIMIT": "KLYS TRON OUTPUT TARGET CONSISTENCY DEGRADE LIMIT (dB)",
                "TS_COHO": "COHO POWER AT A1J4 (dBm)",
                "H_TS_CW": "AME HORIZONTAL TEST SIGNAL POWER (dBm)",
                "TS_RF_SP": "RF DRIVE TEST SIGNAL SHORT PULSE AT 3A5J4 (dBm)",
                "TS_RF_LP": "RF DRIVE TEST SIGNAL LONG PULSE AT 3A5J4 (dBm)",
                "TS_STALO": "STALO POWER AT A1J2 (dBm)",
                "AME_H_NOISE_ENR": "AME NOISE SOURCE HORIZONTAL EXCESS NOISE RATIO (dB)",
                "XMTR_PEAK_PWR_HIGH_LIMIT": "MAXIMUM TRANSMITTER PEAK POWER ALARM LEVEL (kW)",
                "XMTR_PEAK_PWR_LOW_LIMIT": "MINIMUM TRANSMITTER PEAK POWER ALARM LEVEL (kW)",
                "H_DBZ0_DELTA_LIMIT": "DIFFERENCE BETWEEN COMPUTED AND TARGET HORIZONTAL DBZ0 LIMIT (dB)",
                "THRESHOLD1": "BYPASS MAP GENERATOR NOISE THRESHOLD (dB)",
                "THRESHOLD2": "BYPASS MAP GENERATOR REJECTION RATIO THRESHOLD (dB)",
                "CLUT_SUPP_DGRAD_LIM": "CLUTTER SUPPRESSION DEGRADE LIMIT (dB)",
                "CLUT_SUPP_MAINT_LIM": "CLUTTER SUPPRESSION MAINTENANCE LIMIT (dB)",
                "RANGE0_VALUE": "TRUE RANGE AT START OF FIRST RANGE BIN (km)",
                "XMTR_PWR_MTR_SCALE": "SCALE FACTOR USED TO CONVERT TRANSMITTER POWER BYTE DATA TO WATTS (W (4))",
                "V_DBZ0_DELTA_LIMIT": "DIFFERENCE BETWEEN COMPUTED AND TARGET VERTICAL DBZ0 LIMIT (dB)",
                "TAR_H_DBZ0_SP": "HORIZONTAL TARGET SYSTEM CALIBRATION (dBZ0) FOR SHORT PULSE (dBZ)",
                "TAR_V_DBZ0_SP": "VERTICAL TARGET SYSTEM CALIBRATION (DBZ0) FOR SHORT PULSE (dBZ)",
                "DELTAPRF": "SITE PRF SET (A=1, B=2, C=3, D=4, E=5)",
                "TAU_SP": "PULSE WIDTH OF TRANSMITTER OUTPUT IN SHORT PULSE (nsec)",
                "TAU_LP": "PULSE WIDTH OF TRANSMITTER OUTPUT IN LONG PULSE (nsec)",
                "NC_DEAD_VALUE": "NUMBER OF 1/4 KM BINS OF CORRUPTED DATA AT END OF SWEEP",
                "TAU_RF_SP": "RF DRIVE PULSE WIDTH IN SHORT PULSE (nsec)",
                "TAU_RF_LP": "RF DRIVE PULSE WIDTH IN LONG PULSE MODE (nsec)",
                "SEG1LIM": "CLUTTER MAP BOUNDARY ELEVATION BETWEEN SEGMENTS 1 & 2 (deg)",
                "SLATSEC": "SITE LATITUDE - SECONDS (s)",
                "SLONSEC": "SITE LONGITUDE - SECONDS (s)",
                "SLATDEG": "SITE LATITUDE - DEGREES (deg)",
                "SLATMIN": "SITE LATITUDE - MINUTES (min)",
                "SLONDEG": "SITE LONGITUDE - DEGREES (deg)",
                "SLONMIN": "SITE LONGITUDE - MINUTES (min)",
                "SLATDIR": "SITE LATITUDE - DIRECTION",
                "SLONDIR": "SITE LONGITUDE - DIRECTION",
                "VCPAT11": "VOLUME COVERAGE PATTERN NUMBER 11 DEFINITION",
                "VCPAT21": "VOLUME COVERAGE PATTERN NUMBER 21 DEFINITION",
                "VCPAT31": "VOLUME COVERAGE PATTERN NUMBER 31 DEFINITION",
                "VCPAT32": "VOLUME COVERAGE PATTERN NUMBER 32 DEFINITION",
                "VCPAT300": "VOLUME COVERAGE PATTERN NUMBER 300 DEFINITION",
                "VCPAT301": "VOLUME COVERAGE PATTERN NUMBER 301 DEFINITION",
                "AZ_CORRECTION_FACTOR": "AZIMUTH BORESIGHT CORRECTION FACTOR (deg)",
                "EL_CORRECTION_FACTOR": "ELEVATION BORESIGHT CORRECTION FACTOR (deg)",
                "SITE_NAME": "SITE NAME DESIGNATION",
                "ANT_MANUAL_SETUP_IELMIN": "MINIMUM ELEVATION ANGLE (deg)",
                "ANT_MANUAL_SETUP_IELMAX": "MAXIMUM ELEVATION ANGLE (deg)",
                "ANT_MANUAL_SETUP_FAZVELMAX": "MAXIMUM AZIMUTH VELOCITY (deg/s)",
                "ANT_MANUAL_SETUP_FELVELMAX": "MAXIMUM ELEVATION VELOCITY (deg/s)",
                "ANT_MANUAL_SETUP_IGND_HGT": "SITE GROUND HEIGHT (ABOVE SEA LEVEL) (m)",
                "ANT_MANUAL_SETUP_IRAD_HGT": "SITE RADAR HEIGHT (ABOVE GROUND) (m)",
                "RVP8NV_IWAVEGUIDE_LENGTH": "WAVEGUIDE LENGTH (m)",
                "VEL_DATA_TOVER": "VELOCITY UNFOLDING OVERLAY THRESHOLD (dB)",
                "WIDTH_DATA_TOVER": "WIDTH UNFOLDING OVERLAY THRESHOLD (dB)",
                "DOPPLER_RANGE_START": "START RANGE FOR FIRST DOPPLER RADIAL (km)",
                "MAX_EL_INDEX": "THE MAXIMUM INDEX FOR THE EL_INDEX PARAMETERS",
                "SEG2LIM": "CLUTTER MAP BOUNDARY ELEVATION BETWEEN SEGMENTS 2 & 3. (deg)",
                "SEG3LIM": "CLUTTER MAP BOUNDARY ELEVATION BETWEEN SEGMENTS 3 & 4. (deg)",
                "SEG4LIM": "CLUTTER MAP BOUNDARY ELEVATION BETWEEN SEGMENTS 4 & 5. (deg)",
                "NBR_EL_SEGMENTS": "NUMBER OF ELEVATION SEGMENTS IN ORDA CLUTTER MAP.",
                "H_NOISE_LONG": "HORIZONTAL RECEIVER NOISE LONG PULSE (dBm)",
                "ANT_NOISE_TEMP": "ANTENNA NOISE TEMPERATURE (K)",
                "H_NOISE_SHORT": "HORIZONTAL RECEIVER NOISE SHORT PULSE (dBm)",
                "H_NOISE_TOLERANCE": "HORIZONTAL RECEIVER NOISE TOLERANCE (dB)",
                "MIN_H_DYN_RANGE": "MINIMUM HORIZONTAL DYNAMIC RANGE (dB)",
                "GEN_INSTALLED": "AUXILIARY GENERATOR INSTALLED (FAA ONLY)",
                "GEN_EXERCISE": "AUXILIARY GENERATOR AUTOMATIC EXERCISE ENABLED (FAA ONLY)",
                "V_NOISE_TOLERANCE": "VERTICAL RECEIVER NOISE TOLERANCE (dB)",
                "MIN_V_DYN_RANGE": "MINIMUM VERTICAL DYNAMIC RANGE (dB)",
                "ZDR_BIAS_DGRAD_LIM": "SYSTEM DIFFERENTIAL REFLECTIVITY BIAS DEGRADE LIMIT (dB)",
                "V_NOISE_LONG": "VERTICAL RECEIVER NOISE LONG PULSE (dBm)",
                "V_NOISE_SHORT": "VERTICAL RECEIVER NOISE SHORT PULSE (dBm)",
                "ZDR_DATA_TOVER": "ZDR UNFOLDING OVERLAY THRESHOLD (dB)",
                "PHI_DATA_TOVER": "PHI UNFOLDING OVERLAY THRESHOLD (dB)",
                "RHO_DATA_TOVER": "RHO UNFOLDING OVERLAY THRESHOLD (dB)",
                "STALO_POWER_DGRAD_LIMIT": "STALO POWER DEGRADE LIMIT (V)",
                "STALO_POWER_MAINT_LIMIT": "STALO POWER MAINTENANCE LIMIT (V)",
                "MIN_H_PWR_SENSE": "MINIMUM HORIZONTAL POWER SENSE (dBm)",
                "MIN_V_PWR_SENSE": "MINIMUM VERTICAL POWER SENSE (dBm)",
                "H_PWR_SENSE_OFFSET": "HORIZONTAL POWER SENSE CALIBRATION OFFSET (dB)",
                "V_PWR_SENSE_OFFSET": "VERTICAL POWER SENSE CALIBRATION OFFSET (dB)",
                "PS_GAIN_REF": "RF PALLET BIAS ERROR (dB)",
                "RF_PALLET_BROAD_LOSS": "RF PALLET BROADBAND LOSS (dB)",
                "ZDR_CHECK_THRESHOLD": "ZDR CHECK FAILURE THRESHOLD (dB)",
                "PHI_CHECK_THRESHOLD": "PHI CHECK FAILURE THRESHOLD (deg)",
                "RHO_CHECK_THRESHOLD": "RHO CHECK FAILURE THRESHOLD (ratio)",
                "AME_PS_TOLERANCE": "AME POWER SUPPLY TOLERANCE (%)",
                "AME_MAX_TEMP": "MAXIMUM AME INTERNAL ALARM TEMPERATURE (deg C)",
                "AME_MIN_TEMP": "MINIMUM AME INTERNAL ALARM TEMPERATURE (deg C)",
                "RCVR_MOD_MAX_TEMP": "MAXIMUM AME RECEIVER MODULE ALARM TEMPERATURE (deg C)",
                "RCVR_MOD_MIN_TEMP": "MINIMUM AME RECEIVER MODULE ALARM TEMPERATURE (deg C)",
                "BITE_MOD_MAX_TEMP": "MAXIMUM AME BITE MODULE ALARM TEMPERATURE (deg C)",
                "BITE_MOD_MIN_TEMP": "MINIMUM AME BITE MODULE ALARM TEMPERATURE (deg C)",
                "DEFAULT_POLARIZATION": "DEFAULT (H+V) MICROWAVE ASSEMBLY PHASE SHIFTER POSITION",
                "TR_LIMIT_DGRAD_LIMIT": "TR LIMITER DEGRADE LIMIT (V)",
                "TR_LIMIT_FAIL_LIMIT": "TR LIMITER FAILURE LIMIT (V)",
                "AME_CURRENT_TOLERANCE": "AME PELTIER CURRENT TOLERANCE (%)",
                "H_ONLY_POLARIZATION": "HORIZONTAL (H ONLY) MICROWAVE ASSEMBLY PHASE SHIFTER POSITION",
                "V_ONLY_POLARIZATION": "VERTICAL (V ONLY) MICROWAVE ASSEMBLY PHASE SHIFTER POSITION",
                "REFLECTOR_BIAS": "ANTENNA REFLECTOR BIAS (dB)",
                "A_MIN_SHELTER_TEMP_WARN": "LOW EQUIPMENT SHELTER TEMPERATURE WARNING LIMIT (deg C)"}

fields = [("ADAP_FILE_NAME", "12s"),
          ("ADAP_FORMAT", "4s"),
          ("ADAP_REVISION", "4s"),
          ("ADAP_DATE", "12s"),
          ("ADAP_TIME", "12s"),
          ("K1", "f"),
          ("AZ_LAT", "f"),
          ("K3", "f"),
          ("EL_LAT", "f"),
          ("PARKAZ", "f"),
          ("PARKEL", "f"),
          ("A_FUEL_CONV_0", "f"),
          ("A_FUEL_CONV_1", "f"),
          ("A_FUEL_CONV_2", "f"),
          ("A_FUEL_CONV_3", "f"),
          ("A_FUEL_CONV_4", "f"),
          ("A_FUEL_CONV_5", "f"),
          ("A_FUEL_CONV_6", "f"),
          ("A_FUEL_CONV_7", "f"),
          ("A_FUEL_CONV_8", "f"),
          ("A_FUEL_CONV_9", "f"),
          ("A_FUEL_CONV_10", "f"),
          ("A_MIN_SHELTER_TEMP", "f"),
          ("A_MAX_SHELTER_TEMP", "f"),
          ("A_MIN_SHELTER_AC_TEMP_DIFF", "f"),
          ("A_MAX_XMTR_AIR_TEMP", "f"),
          ("A_MAX_RAD_TEMP", "f"),
          ("A_MAX_RAD_TEMP_RISE", "f"),
          ("PED_28V_REG_LIM", "f"),
          ("PED_5V_REG_LIM", "f"),
          ("PED_15V_REG_LIM", "f"),
          ("A_MIN_GEN_ROOM_TEMP", "f"),
          ("A_MAX_GEN_ROOM_TEMP", "f"),
          ("DAU_5V_REG_LIM", "f"),
          ("DAU_15V_REG_LIM", "f"),
          ("DAU_28V_REG_LIM", "f"),
          ("EN_5V_REG_LIM", "f"),
          ("EN_5V_NOM_VOLTS", "f"),
          ("RPG_CO_LOCATED", "4s"),
          ("SPEC_FILTER_INSTALLED", "4s"),
          ("TPS_INSTALLED", "4s"),
          ("RMS_INSTALLED", "4s"),
          ("A_HVDL_TST_INT", "L"),
          ("A_RPG_LT_INT", "L"),
          ("A_MIN_STAB_UTIL_PWR_TIME", "L"),
          ("A_GEN_AUTO_EXER_INTERVAL", "L"),
          ("A_UTIL_PWR_SW_REQ_INTERVAL", "L"),
          ("A_LOW_FUEL_LEVEL", "f"),
          ("CONFIG_CHAN_NUMBER", "L"),
          ("A_RPG_LINK_TYPE", "L"),
          ("REDUNDANT_CHAN_CONFIG", "L"),
          ("ATTEN_TABLE_0", "f"),
          ("ATTEN_TABLE_1", "f"),
          ("ATTEN_TABLE_2", "f"),
          ("ATTEN_TABLE_3", "f"),
          ("ATTEN_TABLE_4", "f"),
          ("ATTEN_TABLE_5", "f"),
          ("ATTEN_TABLE_6", "f"),
          ("ATTEN_TABLE_7", "f"),
          ("ATTEN_TABLE_8", "f"),
          ("ATTEN_TABLE_9", "f"),
          ("ATTEN_TABLE_10", "f"),
          ("ATTEN_TABLE_11", "f"),
          ("ATTEN_TABLE_12", "f"),
          ("ATTEN_TABLE_13", "f"),
          ("ATTEN_TABLE_14", "f"),
          ("ATTEN_TABLE_15", "f"),
          ("ATTEN_TABLE_16", "f"),
          ("ATTEN_TABLE_17", "f"),
          ("ATTEN_TABLE_18", "f"),
          ("ATTEN_TABLE_19", "f"),
          ("ATTEN_TABLE_20", "f"),
          ("ATTEN_TABLE_21", "f"),
          ("ATTEN_TABLE_22", "f"),
          ("ATTEN_TABLE_23", "f"),
          ("ATTEN_TABLE_24", "f"),
          ("ATTEN_TABLE_25", "f"),
          ("ATTEN_TABLE_26", "f"),
          ("ATTEN_TABLE_27", "f"),
          ("ATTEN_TABLE_28", "f"),
          ("ATTEN_TABLE_29", "f"),
          ("ATTEN_TABLE_30", "f"),
          ("ATTEN_TABLE_31", "f"),
          ("ATTEN_TABLE_32", "f"),
          ("ATTEN_TABLE_33", "f"),
          ("ATTEN_TABLE_34", "f"),
          ("ATTEN_TABLE_35", "f"),
          ("ATTEN_TABLE_36", "f"),
          ("ATTEN_TABLE_37", "f"),
          ("ATTEN_TABLE_38", "f"),
          ("ATTEN_TABLE_39", "f"),
          ("ATTEN_TABLE_40", "f"),
          ("ATTEN_TABLE_41", "f"),
          ("ATTEN_TABLE_42", "f"),
          ("ATTEN_TABLE_43", "f"),
          ("ATTEN_TABLE_44", "f"),
          ("ATTEN_TABLE_45", "f"),
          ("ATTEN_TABLE_46", "f"),
          ("ATTEN_TABLE_47", "f"),
          ("ATTEN_TABLE_48", "f"),
          ("ATTEN_TABLE_49", "f"),
          ("ATTEN_TABLE_50", "f"),
          ("ATTEN_TABLE_51", "f"),
          ("ATTEN_TABLE_52", "f"),
          ("ATTEN_TABLE_53", "f"),
          ("ATTEN_TABLE_54", "f"),
          ("ATTEN_TABLE_55", "f"),
          ("ATTEN_TABLE_56", "f"),
          ("ATTEN_TABLE_57", "f"),
          ("ATTEN_TABLE_58", "f"),
          ("ATTEN_TABLE_59", "f"),
          ("ATTEN_TABLE_60", "f"),
          ("ATTEN_TABLE_61", "f"),
          ("ATTEN_TABLE_62", "f"),
          ("ATTEN_TABLE_63", "f"),
          ("ATTEN_TABLE_64", "f"),
          ("ATTEN_TABLE_65", "f"),
          ("ATTEN_TABLE_66", "f"),
          ("ATTEN_TABLE_67", "f"),
          ("ATTEN_TABLE_68", "f"),
          ("ATTEN_TABLE_69", "f"),
          ("ATTEN_TABLE_70", "f"),
          ("ATTEN_TABLE_71", "f"),
          ("ATTEN_TABLE_72", "f"),
          ("ATTEN_TABLE_73", "f"),
          ("ATTEN_TABLE_74", "f"),
          ("ATTEN_TABLE_75", "f"),
          ("ATTEN_TABLE_76", "f"),
          ("ATTEN_TABLE_77", "f"),
          ("ATTEN_TABLE_78", "f"),
          ("ATTEN_TABLE_79", "f"),
          ("ATTEN_TABLE_80", "f"),
          ("ATTEN_TABLE_81", "f"),
          ("ATTEN_TABLE_82", "f"),
          ("ATTEN_TABLE_83", "f"),
          ("ATTEN_TABLE_84", "f"),
          ("ATTEN_TABLE_85", "f"),
          ("ATTEN_TABLE_86", "f"),
          ("ATTEN_TABLE_87", "f"),
          ("ATTEN_TABLE_88", "f"),
          ("ATTEN_TABLE_89", "f"),
          ("ATTEN_TABLE_90", "f"),
          ("ATTEN_TABLE_91", "f"),
          ("ATTEN_TABLE_92", "f"),
          ("ATTEN_TABLE_93", "f"),
          ("ATTEN_TABLE_94", "f"),
          ("ATTEN_TABLE_95", "f"),
          ("ATTEN_TABLE_96", "f"),
          ("ATTEN_TABLE_97", "f"),
          ("ATTEN_TABLE_98", "f"),
          ("ATTEN_TABLE_99", "f"),
          ("ATTEN_TABLE_100", "f"),
          ("ATTEN_TABLE_101", "f"),
          ("ATTEN_TABLE_102", "f"),
          ("ATTEN_TABLE_103", "f"),
          (None, "24x"),
          ("PATH_LOSSES_7", "f"),
          (None, "12x"),
          ("PATH_LOSSES_11", "f"),
          (None, "4x"),
          ("PATH_LOSSES_13", "f"),
          ("PATH_LOSSES_14", "f"),
          (None, "16x"),
          ("PATH_LOSSES_19", "f"),
          (None, "32x"),
          ("PATH_LOSSES_28", "f"),
          (None, "4x"),
          ("PATH_LOSSES_30", "f"),
          ("PATH_LOSSES_31", "f"),
          ("PATH_LOSSES_32", "f"),
          ("PATH_LOSSES_33", "f"),
          ("PATH_LOSSES_34", "f"),
          ("PATH_LOSSES_35", "f"),
          ("PATH_LOSSES_36", "f"),
          ("PATH_LOSSES_37", "f"),
          ("PATH_LOSSES_38", "f"),
          ("PATH_LOSSES_39", "f"),
          ("PATH_LOSSES_40", "f"),
          ("PATH_LOSSES_41", "f"),
          ("PATH_LOSSES_42", "f"),
          ("PATH_LOSSES_43", "f"),
          ("PATH_LOSSES_44", "f"),
          ("PATH_LOSSES_45", "f"),
          ("PATH_LOSSES_46", "f"),
          ("PATH_LOSSES_47", "f"),
          ("H_COUPLER_CW_LOSS", "f"),
          ("V_COUPLER_XMT_LOSS", "f"),
          ("PATH_LOSSES_50", "f"),
          (None, "4x"),
          ("PATH_LOSSES_52", "f"),
          ("V_COUPLER_CW_LOSS", "f"),
          ("PATH_LOSSES_54", "f"),
          (None, "4x"),
          (None, "4x"),
          (None, "4x"),
          ("PATH_LOSSES_58", "f"),
          ("PATH_LOSSES_59", "f"),
          ("PATH_LOSSES_60", "f"),
          ("PATH_LOSSES_61", "f"),
          (None, "4x"),
          ("PATH_LOSSES_63", "f"),
          ("PATH_LOSSES_64", "f"),
          ("PATH_LOSSES_65", "f"),
          ("PATH_LOSSES_66", "f"),
          ("PATH_LOSSES_67", "f"),
          ("PATH_LOSSES_68", "f"),
          ("PATH_LOSSES_69", "f"),
          ("CHAN_CAL_DIFF", "f"),
          (None, "4x"),
          ("LOG_AMP_FACTOR_1", "f"),
          ("LOG_AMP_FACTOR_2", "f"),
          ("V_TS_CW", "f"),
          ("RNSCALE_0", "f"),
          ("RNSCALE_1", "f"),
          ("RNSCALE_2", "f"),
          ("RNSCALE_3", "f"),
          ("RNSCALE_4", "f"),
          ("RNSCALE_5", "f"),
          ("RNSCALE_6", "f"),
          ("RNSCALE_7", "f"),
          ("RNSCALE_8", "f"),
          ("RNSCALE_9", "f"),
          ("RNSCALE_10", "f"),
          ("RNSCALE_11", "f"),
          ("RNSCALE_12", "f"),
          ("ATMOS_0", "f"),
          ("ATMOS_1", "f"),
          ("ATMOS_2", "f"),
          ("ATMOS_3", "f"),
          ("ATMOS_4", "f"),
          ("ATMOS_5", "f"),
          ("ATMOS_6", "f"),
          ("ATMOS_7", "f"),
          ("ATMOS_8", "f"),
          ("ATMOS_9", "f"),
          ("ATMOS_10", "f"),
          ("ATMOS_11", "f"),
          ("ATMOS_12", "f"),
          ("EL_INDEX_0", "f"),
          ("EL_INDEX_1", "f"),
          ("EL_INDEX_2", "f"),
          ("EL_INDEX_3", "f"),
          ("EL_INDEX_4", "f"),
          ("EL_INDEX_5", "f"),
          ("EL_INDEX_6", "f"),
          ("EL_INDEX_7", "f"),
          ("EL_INDEX_8", "f"),
          ("EL_INDEX_9", "f"),
          ("EL_INDEX_10", "f"),
          ("EL_INDEX_11", "f"),
          ("TFREQ_MHZ", "L"),
          ("BASE_DATA_TCN", "f"),
          ("REFL_DATA_TOVER", "f"),
          ("TAR_H_DBZ0_LP", "f"),
          ("TAR_V_DBZ0_LP", "f"),
          ("INIT_PHI_DP", "L"),
          ("NORM_INIT_PHI_DP", "L"),
          ("LX_LP", "f"),
          ("LX_SP", "f"),
          ("METEOR_PARAM", "f"),
          ("BEAMWIDTH", "f"),
          ("ANTENNA_GAIN", "f"),
          (None, "4x"),
          ("VEL_MAINT_LIMIT", "f"),
          ("WTH_MAINT_LIMIT", "f"),
          ("VEL_DEGRAD_LIMIT", "f"),
          ("WTH_DEGRAD_LIMIT", "f"),
          ("H_NOISETEMP_DGRAD_LIMIT", "f"),
          ("H_NOISETEMP_MAINT_LIMIT", "f"),
          ("V_NOISETEMP_DGRAD_LIMIT", "f"),
          ("V_NOISETEMP_MAINT_LIMIT", "f"),
          ("KLY_DEGRADE_LIMIT", "f"),
          ("TS_COHO", "f"),
          ("H_TS_CW", "f"),
          ("TS_RF_SP", "f"),
          ("TS_RF_LP", "f"),
          ("TS_STALO", "f"),
          ("AME_H_NOISE_ENR", "f"),
          ("XMTR_PEAK_PWR_HIGH_LIMIT", "f"),
          ("XMTR_PEAK_PWR_LOW_LIMIT", "f"),
          ("H_DBZ0_DELTA_LIMIT", "f"),
          ("THRESHOLD1", "f"),
          ("THRESHOLD2", "f"),
          ("CLUT_SUPP_DGRAD_LIM", "f"),
          ("CLUT_SUPP_MAINT_LIM", "f"),
          ("RANGE0_VALUE", "f"),
          ("XMTR_PWR_MTR_SCALE", "f"),
          ("V_DBZ0_DELTA_LIMIT", "f"),
          ("TAR_H_DBZ0_SP", "f"),
          ("TAR_V_DBZ0_SP", "f"),
          ("DELTAPRF", "L"),
          (None, "4x"),
          (None, "4x"),
          ("TAU_SP", "L"),
          ("TAU_LP", "L"),
          ("NC_DEAD_VALUE", "L"),
          ("TAU_RF_SP", "L"),
          ("TAU_RF_LP", "L"),
          ("SEG1LIM", "f"),
          ("SLATSEC", "f"),
          ("SLONSEC", "f"),
          (None, "4x"),
          ("SLATDEG", "L"),
          ("SLATMIN", "L"),
          ("SLONDEG", "L"),
          ("SLONMIN", "L"),
          ("SLATDIR", "4s"),
          ("SLONDIR", "4s"),
          (None, "4x"),
          ("VCPAT11", "1172s"),
          ("VCPAT21", "1172s"),
          ("VCPAT31", "1172s"),
          ("VCPAT32", "1172s"),
          ("VCPAT300", "1172s"),
          ("VCPAT301", "1172s"),
          ("AZ_CORRECTION_FACTOR", "f"),
          ("EL_CORRECTION_FACTOR", "f"),
          ("SITE_NAME", "4s"),
          ("ANT_MANUAL_SETUP_IELMIN", "l"),
          ("ANT_MANUAL_SETUP_IELMAX", "L"),
          ("ANT_MANUAL_SETUP_FAZVELMAX", "L"),
          ("ANT_MANUAL_SETUP_FELVELMAX", "L"),
          ("ANT_MANUAL_SETUP_IGND_HGT", "L"),
          ("ANT_MANUAL_SETUP_IRAD_HGT", "L"),
          (None, "300x"),
          ("RVP8NV_IWAVEGUIDE_LENGTH", "L"),
          (None, "44x"),
          ("VEL_DATA_TOVER", "f"),
          ("WIDTH_DATA_TOVER", "f"),
          (None, "12x"),
          ("DOPPLER_RANGE_START", "f"),
          ("MAX_EL_INDEX", "L"),
          ("SEG2LIM", "f"),
          ("SEG3LIM", "f"),
          ("SEG4LIM", "f"),
          ("NBR_EL_SEGMENTS", "L"),
          ("H_NOISE_LONG", "f"),
          ("ANT_NOISE_TEMP", "f"),
          ("H_NOISE_SHORT", "f"),
          ("H_NOISE_TOLERANCE", "f"),
          ("MIN_H_DYN_RANGE", "f"),
          ("GEN_INSTALLED", "4s"),
          ("GEN_EXERCISE", "4s"),
          ("V_NOISE_TOLERANCE", "f"),
          ("MIN_V_DYN_RANGE", "f"),
          ("ZDR_BIAS_DGRAD_LIM", "f"),
          (None, "4x"),
          (None, "4x"),
          (None, "8x"),
          ("V_NOISE_LONG", "f"),
          ("V_NOISE_SHORT", "f"),
          ("ZDR_DATA_TOVER", "f"),
          ("PHI_DATA_TOVER", "f"),
          ("RHO_DATA_TOVER", "f"),
          ("STALO_POWER_DGRAD_LIMIT", "f"),
          ("STALO_POWER_MAINT_LIMIT", "f"),
          ("MIN_H_PWR_SENSE", "f"),
          ("MIN_V_PWR_SENSE", "f"),
          ("H_PWR_SENSE_OFFSET", "f"),
          ("V_PWR_SENSE_OFFSET", "f"),
          ("PS_GAIN_REF", "f"),
          ("RF_PALLET_BROAD_LOSS", "f"),
          ("ZDR_CHECK_THRESHOLD", "f"),
          ("PHI_CHECK_THRESHOLD", "f"),
          ("RHO_CHECK_THRESHOLD", "f"),
          (None, "52x"),
          ("AME_PS_TOLERANCE", "f"),
          ("AME_MAX_TEMP", "f"),
          ("AME_MIN_TEMP", "f"),
          ("RCVR_MOD_MAX_TEMP", "f"),
          ("RCVR_MOD_MIN_TEMP", "f"),
          ("BITE_MOD_MAX_TEMP", "f"),
          ("BITE_MOD_MIN_TEMP", "f"),
          ("DEFAULT_POLARIZATION", "L"),
          ("TR_LIMIT_DGRAD_LIMIT", "f"),
          ("TR_LIMIT_FAIL_LIMIT", "f"),
          (None, "4x"),
          (None, "4x"),
          ("AME_CURRENT_TOLERANCE", "f"),
          ("H_ONLY_POLARIZATION", "L"),
          ("V_ONLY_POLARIZATION", "L"),
          (None, "8x"),
          ("REFLECTOR_BIAS", "f"),
          ("A_MIN_SHELTER_TEMP_WARN", "f"),
          (None, "432x")]



# Copyright (c) 2015,2017,2019 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
r"""Module to provide unit support.

This makes use of the ``pint`` library and sets up the default settings
for good temperature support.

See Also: :doc:`Working with Units </tutorials/unit_tutorial>`.

Attributes
----------
units : :class: `pint.UnitRegistry()`
    The unit registry used throughout the package. Any use of units in MetPy should
    import this registry and use it to grab units.

"""
import contextlib
import functools
from inspect import Parameter, signature
import logging
import re
import warnings

import numpy as np
import pint

from .package_tools import Exporter

exporter = Exporter(globals())

log = logging.getLogger(__name__)

UndefinedUnitError = pint.UndefinedUnitError
DimensionalityError = pint.DimensionalityError

_base_unit_of_dimensionality = {
    '[pressure]': 'Pa',
    '[temperature]': 'K',
    '[dimensionless]': '',
    '[length]': 'm',
    '[speed]': 'm s**-1'
}


def _fix_udunits_powers(string):
    """Replace UDUNITS-style powers (m2 s-2) with exponent symbols (m**2 s**-2)."""
    return _UDUNIT_POWER.sub('**', string)


def _fix_udunits_div(string):
    return 's**-1' if string == '/s' else string


# Fix UDUNITS-style powers, percent signs, and ill-defined units
_UDUNIT_POWER = re.compile(r'(?<=[A-Za-z\)])(?![A-Za-z\)])'
                           r'(?<![0-9\-][eE])(?<![0-9\-])(?=[0-9\-])')
_unit_preprocessors = [_fix_udunits_powers, lambda string: string.replace('%', 'percent'),
                       _fix_udunits_div]


@exporter.export
def setup_registry(reg):
    """Set up a given registry with MetPy's default tweaks and settings."""
    reg.autoconvert_offset_to_baseunit = True

    # For Pint 0.18.0, need to deal with the fact that the wrapper isn't forwarding on setting
    # the attribute.
    with contextlib.suppress(AttributeError):
        reg.get().autoconvert_offset_to_baseunit = True

    for pre in _unit_preprocessors:
        if pre not in reg.preprocessors:
            reg.preprocessors.append(pre)

    # Add a percent unit if it's not already present, it was added in 0.21
    if 'percent' not in reg:
        reg.define('percent = 0.01 = %')

    # Define commonly encountered units not defined by pint
    reg.define('degrees_north = degree = degrees_N = degreesN = degree_north = degree_N '
               '= degreeN')
    reg.define('degrees_east = degree = degrees_E = degreesE = degree_east = degree_E '
               '= degreeE')
    reg.define('dBz = 1e-18 m^3; logbase: 10; logfactor: 10 = dBZ')

    # Alias geopotential meters (gpm) to just meters
    reg.define('@alias meter = gpm')

    # Enable pint's built-in matplotlib support
    reg.setup_matplotlib()

    return reg


# Make our modifications using pint's application registry--which allows us to better
# interoperate with other libraries using Pint.
units = setup_registry(pint.get_application_registry())

# Silence UnitStrippedWarning
warnings.simplefilter('ignore', category=pint.UnitStrippedWarning)


@exporter.export
def pandas_dataframe_to_unit_arrays(df, column_units=None):
    """Attach units to data in pandas dataframes and return quantities.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Data in pandas dataframe.

    column_units : dict
        Dictionary of units to attach to columns of the dataframe. Overrides
        the units attribute if it is attached to the dataframe.

    Returns
    -------
        Dictionary containing `Quantity` instances with keys corresponding to the dataframe
        column names.

    """
    if not column_units:
        try:
            column_units = df.units
        except AttributeError:
            raise ValueError('No units attribute attached to pandas '
                             'dataframe and col_units not given.') from None

    # Iterate through columns attaching units if we have them, if not, don't touch it
    res = {}
    for column in df:
        if column in column_units and column_units[column]:
            res[column] = units.Quantity(df[column].values, column_units[column])
        else:
            res[column] = df[column].values
    return res


def is_quantity(*args):
    """Check whether an instance is a quantity."""
    return all(isinstance(a, pint.Quantity) for a in args)


def concatenate(arrs, axis=0):
    r"""Concatenate multiple values into a new quantity.

    This is essentially a scalar-/masked array-aware version of `numpy.concatenate`. All items
    must be able to be converted to the same units. If an item has no units, it will be given
    those of the rest of the collection, without conversion. The first units found in the
    arguments is used as the final output units.

    Parameters
    ----------
    arrs : Sequence[pint.Quantity or numpy.ndarray]
        The items to be joined together

    axis : int, optional
        The array axis along which to join the arrays. Defaults to 0 (the first dimension)

    Returns
    -------
    `pint.Quantity`
        New container with the value passed in and units corresponding to the first item.

    """
    dest = 'dimensionless'
    for a in arrs:
        if hasattr(a, 'units'):
            dest = a.units
            break

    data = []
    for a in arrs:
        if hasattr(a, 'to'):
            a = a.to(dest).magnitude
        data.append(np.atleast_1d(a))

    # Use masked array concatenate to ensure masks are preserved, but convert to an
    # array if there are no masked values.
    data = np.ma.concatenate(data, axis=axis)
    if not np.any(data.mask):
        data = np.asarray(data)

    return units.Quantity(data, dest)


def masked_array(data, data_units=None, **kwargs):
    """Create a :class:`numpy.ma.MaskedArray` with units attached.

    This is a thin wrapper around :class:`numpy.ma.MaskedArray` that ensures that
    units are properly attached to the result (otherwise units are silently lost). Units
    are taken from the ``data_units`` argument, or if this is ``None``, the units on ``data``
    are used.

    Parameters
    ----------
    data : array-like
        The source data. If ``data_units`` is `None`, this should be a `pint.Quantity` with
        the desired units.
    data_units : str or `pint.Unit`, optional
        The units for the resulting `pint.Quantity`
    kwargs
        Arbitrary keyword arguments passed to `numpy.ma.masked_array`, optional

    Returns
    -------
    `pint.Quantity`

    """
    if data_units is None:
        data_units = data.units
        data = data.magnitude
    return units.Quantity(np.ma.masked_array(data, **kwargs), data_units)


def _mutate_arguments(bound_args, check_type, mutate_arg):
    """Handle adjusting bound arguments.

    Calls ``mutate_arg`` on every argument, including those passed as ``*args``, if they are
    of type ``check_type``.
    """
    for arg_name, arg_val in bound_args.arguments.items():
        if isinstance(arg_val, check_type):
            bound_args.arguments[arg_name] = mutate_arg(arg_val, arg_name)

    if isinstance(bound_args.arguments.get('args'), tuple):
        bound_args.arguments['args'] = tuple(
            mutate_arg(arg_val, '(unnamed)') if isinstance(arg_val, check_type) else arg_val
            for arg_val in bound_args.arguments['args'])


def _check_argument_units(args, defaults, dimensionality):
    """Yield arguments with improper dimensionality."""
    for arg, val in args.items():
        # Get the needed dimensionality (for printing) as well as cached, parsed version
        # for this argument.
        try:
            need, parsed = dimensionality[arg]
        except KeyError:
            # Argument did not have units specified in decorator
            continue

        if arg in defaults and (defaults[arg] is not None or val is None):
            check = val == defaults[arg]
            if np.all(check):
                continue

        # See if the value passed in is appropriate
        try:
            if val.dimensionality != parsed:
                yield arg, val, val.units, need
        # No dimensionality
        except AttributeError:
            # If this argument is dimensionless, don't worry
            if parsed != '':
                yield arg, val, 'none', need


def _get_changed_versions(docstring):
    """Find the most recent version in which the docs say a function changed."""
    matches = re.findall(r'.. versionchanged:: ([\d.]+)', docstring)
    return matches


def _check_units_outer_helper(func, *args, **kwargs):
    """Get dims and defaults from function signature and specified dimensionalities."""
    # Match the signature of the function to the arguments given to the decorator
    sig = signature(func)
    bound_units = sig.bind_partial(*args, **kwargs)

    # Convert our specified dimensionality (e.g. "[pressure]") to one used by
    # pint directly (e.g. "[mass] / [length] / [time]**2). This is for both efficiency
    # reasons and to ensure that problems with the decorator are caught at import,
    # rather than runtime.
    dims = {name: (orig, units.get_dimensionality(orig.replace('dimensionless', '')))
            for name, orig in bound_units.arguments.items()}

    defaults = {name: sig.parameters[name].default for name in sig.parameters
                if sig.parameters[name].default is not Parameter.empty}

    return sig, dims, defaults


def _check_units_inner_helper(func, sig, defaults, dims, *args, **kwargs):
    """Check bound arguments for unit correctness."""
    # Match all passed in value to their proper arguments so we can check units
    bound_args = sig.bind(*args, **kwargs)
    bad = list(_check_argument_units(bound_args.arguments, defaults, dims))

    # If there are any bad units, emit a proper error message making it clear
    # what went wrong.
    if bad:
        msg = f'`{func.__name__}` given arguments with incorrect units: '
        msg += ', '.join(
            f'`{arg}` requires "{req}" but given "{given}"' for arg, _, given, req in bad
        )
        if 'none' in msg:
            if any(isinstance(x, np.ma.core.MaskedArray) for _, x, _, _ in bad):
                msg += ('\nA masked array `m` can be assigned a unit as follows:\n'
                        '    from metpy.units import units\n'
                        '    m = units.Quantity(m, "m/s")')
            else:
                msg += ('\nA xarray DataArray or numpy array `x` can be assigned a unit as '
                        'follows:\n'
                        '    from metpy.units import units\n'
                        '    x = x * units("m/s")')
            msg += ('\nFor more information see the Units Tutorial: '
                    'https://unidata.github.io/MetPy/latest/tutorials/unit_tutorial.html')

        # If function has changed, mention that fact
        if func.__doc__:
            changed_versions = _get_changed_versions(func.__doc__)
            if changed_versions:
                msg = (
                    f'This function changed in version(s) {changed_versions}--double check '
                    'that the function is being called properly.\n'
                ) + msg
        raise ValueError(msg)

    # Return the bound arguments for reuse
    return bound_args


def check_units(*units_by_pos, **units_by_name):
    """Create a decorator to check units of function arguments."""
    def dec(func):
        sig, dims, defaults = _check_units_outer_helper(func, *units_by_pos, **units_by_name)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _check_units_inner_helper(func, sig, defaults, dims, *args, **kwargs)
            return func(*args, **kwargs)

        return wrapper
    return dec


def process_units(
    input_dimensionalities,
    output_dimensionalities,
    output_to=None,
    ignore_inputs_for_output=None
):
    """Wrap a non-Quantity-using function in base units to fully handle units."""
    def dec(func):
        sig, dims, defaults = _check_units_outer_helper(func, **input_dimensionalities)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            bound_args = _check_units_inner_helper(func, sig, defaults, dims, *args, **kwargs)

            # Determine unit(s) with which to wrap output (first, since we mutate the bound
            # args)
            if isinstance(output_dimensionalities, tuple):
                multiple_output = True
                outputs = output_dimensionalities
            else:
                multiple_output = False
                outputs = (output_dimensionalities,)
            output_control = []
            for i, output in enumerate(outputs):
                convert_to = (
                    output_to if not multiple_output or output_to is None else output_to[i]
                )
                # Find matching input, if it exists
                if convert_to is None:
                    for name, (this_dim, _) in dims.items():
                        if (
                            this_dim == output
                            and (
                                ignore_inputs_for_output is None
                                or name not in ignore_inputs_for_output
                            )
                        ):
                            try:
                                convert_to = bound_args.arguments[name].units
                            except AttributeError:
                                # We don't have units, so given prior check, is dimensionless
                                convert_to = ''
                            break

                output_control.append((_base_unit_of_dimensionality[output], convert_to))

            # Convert all inputs as specified, assuming dimensionality is fine based on above
            _mutate_arguments(bound_args, units.Quantity, lambda val, _: val.to_base_units().m)

            # Evaluate inner calculation
            result = func(*bound_args.args, **bound_args.kwargs)

            # Wrap output
            if multiple_output:
                wrapped_result = []
                for this_result, this_output_control in zip(result, output_control):
                    q = units.Quantity(this_result, this_output_control[0])
                    if this_output_control[1] is not None:
                        q = q.to(this_output_control[1])
                    wrapped_result.append(q)
                return tuple(wrapped_result)
            else:
                q = units.Quantity(result, output_control[0][0])
                if output_control[0][1] is not None:
                    q = q.to(output_control[0][1])
                return q

        # Attach the unwrapped func for internal use
        wrapper._nounit = func

        return wrapper
    return dec
